"""Generic extension of mwclient.site, mainly to provide multi-slot page handling and
caching OpenSemanticLab specific features are located in osw.core.OSW
"""

import json
import os
import shutil
import urllib
import warnings
import xml.etree.ElementTree as et
from copy import deepcopy
from datetime import datetime
from io import StringIO
from pathlib import Path
from pprint import pprint
from time import sleep
from typing import Any, Dict, List, Optional, Union
from warnings import warn

import mwclient
import pyld
import requests
from jsonpath_ng.ext import parse
from mwclient.page import Page as MwPage
from pydantic.v1 import FilePath
from typing_extensions import deprecated

import osw.model.page_package as package
import osw.utils.util as ut
import osw.wiki_tools as wt
from osw.auth import CredentialManager
from osw.model.static import OswBaseModel
from osw.utils.regex_pattern import REGEX_PATTERN_LIB
from osw.utils.util import parallelize
from osw.utils.wiki import get_osw_id

# Constants
SLOTS = {
    "main": {"content_model": "wikitext", "content_template": ""},
    "header": {"content_model": "wikitext", "content_template": ""},
    "footer": {"content_model": "wikitext", "content_template": ""},
    "jsondata": {"content_model": "json", "content_template": ""},
    "jsonschema": {"content_model": "json", "content_template": ""},
    "template": {"content_model": "wikitext", "content_template": ""},
    "header_template": {"content_model": "wikitext", "content_template": ""},
    "footer_template": {"content_model": "wikitext", "content_template": ""},
    "data_template": {"content_model": "wikitext", "content_template": ""},
    "schema_template": {"content_model": "wikitext", "content_template": ""},
}


# Classes
class WtSite:
    """A wrapper class of mwclient.Site, mainly to provide multi-slot page handling and
    caching"""

    class WtSiteConfig(OswBaseModel):
        """The configuration for a WtSite instance"""

        iri: str
        """The IRI of the wiki site, typically the domain"""
        cred_mngr: CredentialManager
        """The CredentialManager to use for authentication"""
        login: Optional[str]
        """The preferred login name when multiple logins are possible (not supported
        yet)"""
        connection_options: Optional[Dict[str, Any]] = None
        """Parameters for the mwclient.Site connection: Additional arguments to be
        passed to the requests.Session.request() method when performing API calls."""

    @deprecated("Use WtSiteConfig instead")
    class WtSiteLegacyConfig(OswBaseModel):
        """The legacy configuration for a WtSite instance"""

        site: mwclient.Site
        """the mwclient.Site to use for communication with the wiki"""

        class Config:
            arbitrary_types_allowed = True

    def __init__(self, config: Union[WtSiteConfig, WtSiteLegacyConfig]):
        """creates a new WtSite instance from a WtSiteConfig

        Parameters
        ----------
        config
            the WtSiteConfig or WtSiteLegacyConfig to create the WtSite from

        """

        scheme = "https"

        if isinstance(config, WtSite.WtSiteLegacyConfig):
            self._site: mwclient.Site = config.site
        else:
            cred = config.cred_mngr.get_credential(
                CredentialManager.CredentialConfig(
                    iri=config.iri, fallback=CredentialManager.CredentialFallback.ask
                )
            )
            if "//" in config.iri:
                scheme = config.iri.split("://")[0]
                config.iri = config.iri.split("://")[1]
            site_args = [config.iri]
            # increase pool_maxsize to improve performance with many requests to the same server
            # see: https://stackoverflow.com/questions/18466079/change-the-connection-pool-size-for-pythons-requests-module-when-in-threading # noqa
            session = requests.Session()
            adapter = requests.adapters.HTTPAdapter(pool_maxsize=50)
            session.mount(scheme + "://", adapter)
            # verify=False is necessary for self-signed / outdated certificates
            site_kwargs = {
                "path": "/w/",
                "scheme": scheme,
                "pool": session,
                "do_init": False,  # do not initialize the site metadata for performance reasons
                "connection_options": {
                    "verify": True,
                },
            }
            if getattr(config, "connection_options", None) is not None:
                # merge connection_options into reqs
                site_kwargs["connection_options"] = {
                    **site_kwargs["connection_options"],
                    **config.connection_options,
                }
                # reqs might be a deprecated alias for "connection_options"
            if isinstance(cred, CredentialManager.UserPwdCredential):
                self._site = mwclient.Site(*site_args, **site_kwargs)
                self._site.login(username=cred.username, password=cred.password)
            elif isinstance(cred, CredentialManager.OAuth1Credential):
                site_kwargs.update(
                    **{
                        "consumer_token": cred.consumer_token,
                        "consumer_secret": cred.consumer_secret,
                        "access_token": cred.access_token,
                        "access_secret": cred.access_secret,
                    }
                )
                self._site = mwclient.Site(*site_args, **site_kwargs)
            else:
                raise ValueError("Unsupported credential type: " + str(type(cred)))
            del cred

        # The page cache is used to store pages that already have been loaded from
        #  the wiki
        self._page_cache = {}
        self._cache_enabled = False

    @property
    def mw_site(self) -> mwclient.client.Site:
        """Returns the mwclient.Site object of the WtSite instance"""
        return self._site

    @classmethod
    @deprecated("Use contructor instead")
    def from_domain(
        cls,
        domain: str = None,
        password_file: Union[str, FilePath] = None,
        credentials: dict = None,
    ):
        """creates a new WtSite instance from a domain and a password file

        Parameters
        ----------
        domain
            the wiki domain, by default None
        password_file
            password file with credentials for the wiki site, by default None
        credentials
            credentials of the wiki site, by default None

        Returns
        -------
            a new WtSite instance
        """
        if credentials is None:
            site = wt.create_site_object(domain, password_file)
        else:
            site = wt.create_site_object(domain, "", credentials)
        return cls(WtSite.WtSiteLegacyConfig(site=site))

    @classmethod
    @deprecated("Use contructor instead")
    def from_credentials(
        cls,
        credentials: Union[Dict[str, Dict[str, str]], str, FilePath],
        key: Union[str, int] = 0,
    ):
        """

        Parameters
        ----------
        credentials:
            A dictionary of credentials or a path to a credentials file, expected to
            contain keys 'username' and 'password'.
        key:
            The key of the credentials-dictionary (domain) or the index of the key.
            By default, the first key is used.

        Returns
        -------

        """

        if isinstance(credentials, str) or isinstance(credentials, Path):
            domains, accounts = wt.read_domains_from_credentials_file(credentials)
        else:
            accounts = credentials
        if isinstance(key, int):
            _domain = list(accounts.keys())[key]
        else:
            _domain = key
        _domain = _domain
        _credentials: Dict = accounts[_domain]
        # todo: research why nested typing doesn't work as expected fo the dictionary
        site = wt.create_site_object(_domain, "", _credentials)
        return cls(WtSite.WtSiteLegacyConfig(site=site))

    class GetPageParam(OswBaseModel):
        titles: Union[str, List[str]]
        """title string or list of title strings of the pages to download"""
        parallel: Optional[bool] = None
        """whether to download the pages in parallel or sequentially
        Defaults to True if more than 5 pages are requested, False otherwise"""
        retries: Optional[int] = 5
        """How often to retry downloading a page if an error occurs"""
        retry_delay_s: Optional[int] = 5
        """Retry delay in seconds"""
        debug: Optional[bool] = False
        """Whether to print debug messages"""
        raise_exception: Optional[bool] = False
        """Whether to raise an exception if an error occurs"""
        raise_warning: Optional[bool] = True
        """Whether to raise a warning if a page does not exist occurs"""

        def __init__(self, **data):
            super().__init__(**data)
            if not isinstance(self.titles, list):
                self.titles = [self.titles]
            if len(self.titles) > 5 and self.parallel is None:
                self.parallel = True
            if self.parallel is None:
                self.parallel = False

    class GetPageResult(OswBaseModel):
        pages: List["WtPage"]
        """List of pages that have been downloaded"""
        errors: List[Exception]
        """List of errors that occurred while downloading pages"""

        class Config:
            arbitrary_types_allowed = True  # allows to use WtPage in type hints

    def get_page(self, param: GetPageParam) -> GetPageResult:
        """Downloads a page or a list of pages from the site.

        Parameters
        ----------
        param:
            GetPageParam object
        """
        max_index = len(param.titles)

        exceptions = []
        pages = []

        def get_page_(title: str, index: int = None):
            retry = 0
            wtpage = None
            while retry < param.retries:
                msg = ""
                if index is not None:
                    msg = f"({index + 1}/{max_index}) "
                try:
                    if self._cache_enabled and title in self._page_cache:
                        wtpage = self._page_cache[title]
                        msg += "Page loaded from cache. "
                    else:
                        wtpage = WtPage(self, title)
                        msg += "Page loaded. "
                        if self._cache_enabled:
                            self._page_cache[title] = wtpage
                    pages.append(wtpage)
                    if not wtpage.exists:
                        if param.raise_warning:
                            warnings.warn(
                                "WARNING: Page with title '{}' does not exist.".format(
                                    title
                                ),
                                RuntimeWarning,
                                3,
                            )
                        # throw argument value exception if page does not exist
                        raise ValueError(f"Page with title '{title}' does not exist.")

                    break
                except Exception as e:
                    exceptions.append(e)
                    msg += str(e)
                    # retry if probably a network error and retry limit not reached
                    if not isinstance(e, ValueError) and retry < param.retries:
                        retry += 1
                        msg = f"Page load failed: '{e}'. Retry ({retry}/{param.retries}). "
                        sleep(5)
                    else:  # last entry -> throw exception
                        retry = param.retries
                        if param.raise_exception:
                            raise e
                print(msg)
            self._clear_cookies()
            return wtpage

        if param.parallel:
            _ = parallelize(get_page_, param.titles, flush_at_end=param.debug)
        else:
            _ = [get_page_(p, i) for i, p in enumerate(param.titles)]

        return self.GetPageResult(pages=pages, errors=exceptions)

    @deprecated("Use get_page instead")
    def get_WtPage(self, title: str = None):
        """Creates a new WtPage object for the given title
           and loads the page from the site if the page already exists.
           Deprecated in favor of WtSite.get_page.

        Parameters
        ----------
        title:
            title of the page to load / create

        """
        result = self.get_page(WtSite.GetPageParam(titles=title))
        return result.pages[0]

    class GetPageContentResult(OswBaseModel):
        contents: dict
        """The content of the pages. Keys are page titles, values are
        content dictionaries"""

        class Config:
            arbitrary_types_allowed = True

    def get_page_content(self, full_page_titles: List[str]) -> GetPageContentResult:
        get_page_res = self.get_page(WtSite.GetPageParam(titles=full_page_titles))
        contents_dict = {}
        for page in get_page_res.pages:
            title = page.title
            slot_contents = {}
            for slot in SLOTS:
                slot_content = page.get_slot_content(slot)
                if slot_content is not None:
                    slot_contents[slot] = slot_content
            contents_dict[title] = slot_contents

        return WtSite.GetPageContentResult(contents=contents_dict)

    def enable_cache(self):
        """Enables the page cache. If the cache is enabled, pages that have been
        downloaded once are stored in memory and are not downloaded again
        """
        self._cache_enabled = True

    def disable_cache(self):
        """Disables the page cache. If the cache is disabled, pages are downloaded
        every time they are requested
        """
        self._cache_enabled = False

    def get_cache_enabled(self):
        """Returns whether the page cache is enabled

        Returns
        -------
        bool
            whether the page cache is enabled
        """
        return self._cache_enabled

    def clear_cache(self):
        """Clears the page cache. All pages that have been downloaded are removed from
        the cache
        """
        del self._page_cache
        self._page_cache = {}

    def _clear_cookies(self):
        # see https://github.com/mwclient/mwclient/issues/221
        for cookie in self._site.connection.cookies:
            if "PostEditRevision" in cookie.name:
                self._site.connection.cookies.clear(
                    cookie.domain, cookie.path, cookie.name
                )

    class SearchParam(wt.SearchParam):
        pass

    def prefix_search(self, text: Union[str, SearchParam]):
        """Send a prefix search request to the site.

        Parameters
        ----------
        text
            The search text or a SearchParam object

        Returns
        -------
            A list of page titles
        """
        return wt.prefix_search(self._site, text)

    def semantic_search(self, query: Union[str, SearchParam]):
        """Send a swm ask query to the site.

        Parameters
        ----------
        query
            The query text (e. g. "[[Category:Entity]]") or a SearchParam object

        Returns
        -------
            A list of page titles
        """
        return wt.semantic_search(self._site, query)

    class ModifySearchResultsParam(OswBaseModel):
        """Todo: should become param of modify_search_results"""

        mode: str
        """The search mode. Either 'prefix' or 'semantic'."""
        query: wt.SearchParam
        """The search query."""
        comment: str = None
        """The comment for the edit."""
        log: bool = False
        """Whether to log changes."""
        dryrun: bool = False
        """if True, no actual changes are made"""

    def modify_search_results(
        self,
        mode: str,
        query: str,
        modify_page,
        limit=None,
        comment=None,
        log=False,
        dryrun=False,
    ):
        """Modifies the search results of a prefix or semantic search in a callback and
        stores the changes.

        Parameters
        ----------
        mode
            The search mode. Either 'prefix' or 'semantic'.
        query
            The search query.
        modify_page
            The callback that modifies the pages.
        limit
            query limit, by default None
        comment
            edit comment, by default None
        log
            log changes, by default False
        dryrun
            if True, no actual changes are made, by default False
        """
        titles = []
        if mode == "prefix":
            titles = wt.prefix_search(self._site, query)
        elif mode == "semantic":
            titles = wt.semantic_search(self._site, query)
        if limit:
            titles = titles[0:limit]
        if log:
            print(f"Found: {titles}")
        for title in titles:
            wtpage = self.get_page(WtSite.GetPageParam(titles=[title])).pages[0]
            modify_page(wtpage)
            if log:
                print(f"\n======= {title} =======")
                print(wtpage._content)
                for slot in wtpage._slots:
                    content = wtpage.get_slot_content(slot)
                    # if isinstance(content, dict): content = json.dumps(content)
                    print(f"   ==== {title}:{slot} ====   ")
                    pprint(content)
                    print("\n")
            if not dryrun:
                wtpage.edit(comment)

    class UploadPageParam(OswBaseModel):
        """Parameter object for upload_page method."""

        pages: Union["WtPage", List["WtPage"]]
        """A WtPage object or a list of WtPage objects."""
        parallel: Optional[bool] = False
        """If True, uploads the pages in parallel."""
        debug: Optional[bool] = False
        """If True, debug messages will be printed."""

        class Config:
            arbitrary_types_allowed = True

        def __init__(self, **data):
            super().__init__(**data)
            if not isinstance(self.pages, list):
                self.pages = [self.pages]
            if len(self.pages) > 5 and self.parallel is None:
                self.parallel = True
            if self.parallel is None:
                self.parallel = False

    def upload_page(
        self,
        param: Union[UploadPageParam, "WtPage", List["WtPage"]],
    ) -> None:
        """Uploads a page or a list of pages to the site.

        Parameters
        ----------
        param:
            UploadPageParam object or a WtPage object or a list of WtPage objects.
        """
        if not isinstance(param, WtSite.UploadPageParam):
            param = WtSite.UploadPageParam(pages=param)

        max_index = len(param.pages)

        def upload_page_(page, index: int = None):
            # Before uploading: Check if the page is uploaded to the WtSite that is
            #  defining this method
            if page.wtSite != self:
                raise AssertionError(
                    f"The WtSite in page '{page.title}' and the "
                    f"WtSite from which this method is called from "
                    f"are not matching!"
                )
            page.edit()

            if index is None:
                print(f"Uploaded page to {page.get_url()}.")
            else:
                print(
                    f"({index + 1}/{max_index}): Uploaded page to " f"{page.get_url()}."
                )

        if param.parallel:
            _ = parallelize(upload_page_, param.pages, flush_at_end=param.debug)
        else:
            _ = [upload_page_(p, i) for i, p in enumerate(param.pages)]

    class CopyPagesParam(OswBaseModel):
        """Configuration to copy several page"""

        source_site: "WtSite"
        """The source site to copy the pages from"""
        existing_pages: Union[str, List[str]]
        """The full page title of the pages on the source site"""
        overwrite: Optional[bool] = False
        """If true, pages will be overwritten if they already exists on the target
        site"""
        parallel: Optional[bool] = None
        """If true, uploads the pages in parallel."""
        comment: Optional[str] = None
        """Edit comment for the page history. If set to none, will be replaced with
        '[bot edit] Copied from {source_site.host}'."""

        class Config:
            arbitrary_types_allowed = True

        def __init__(self, **data):
            super().__init__(**data)
            if not isinstance(self.existing_pages, list):
                self.existing_pages = [self.existing_pages]
            if len(self.existing_pages) > 5 and self.parallel is None:
                self.parallel = True
            if self.parallel is None:
                self.parallel = False

    def copy_pages(self, param: CopyPagesParam):
        """Copies pages from a source site to this (target) site."""

        def copy_single_page(content_dict: dict):
            title = list(content_dict.keys())[0]
            wtpage = WtPage(wtSite=self, title=title)
            return wtpage.copy(
                config=WtPage.CopyPageConfig(
                    source_site=param.source_site,
                    existing_page=title,
                    overwrite=param.overwrite,
                    comment=param.comment,
                )
            )

        page_contents = param.source_site.get_page_content(param.existing_pages)
        content_list = [{key: value} for key, value in page_contents.contents.items()]

        if param.parallel:
            return ut.parallelize(
                copy_single_page,
                content_list,
                flush_at_end=True,
            )
        else:
            return [copy_single_page(content) for content in content_list]

    class DeletePageParam(OswBaseModel):
        page: Union["WtPage", List["WtPage"], str, List[str]]
        comment: Optional[str] = None
        debug: Optional[bool] = True
        """If True, debug messages will be printed."""
        parallel: Optional[bool] = None
        """If true, processes the pages in parallel."""

        class Config:
            arbitrary_types_allowed = True

        def __init__(self, **data):
            super().__init__(**data)
            if not isinstance(self.page, list):
                self.page = [self.page]
            if not self.comment:
                self.comment = "Deleted via osw-python"
            if len(self.page) > 5 and self.parallel is None:
                self.parallel = True

    def delete_page(
        self,
        param: Union["WtPage", List["WtPage"], str, List[str], DeletePageParam],
        comment: str = None,
    ):
        """Deletes a page or a list of pages from the site.

        Parameters
        ----------
        param:
            DeletePageParam object
        comment:
            The comment to leave when deleting the page
        """
        if not isinstance(param, WtSite.DeletePageParam):
            param = WtSite.DeletePageParam(page=param)
        if comment:
            param.comment = comment
        pages = []
        print("Getting pages to delete ...")
        for page in param.page:
            if isinstance(page, str):
                string_in_pages = True
                try:
                    page = self._site.pages[page]
                    pages.append(page)
                except Exception as e:
                    warn(
                        f"Page '{page}' could not be added to the list of "
                        f"to-be-deleted pages. The following Exception occurred:\n{e}"
                    )
            else:  # page is a WtPage object
                pages.append(page)

        if string_in_pages:
            self._site.tokens["csrf"] = self._site.get_token("csrf")

        def delete_single_page(page_: Union["WtPage", MwPage], comment: str):
            if isinstance(page_, WtPage):
                return page_.delete(comment=comment)
            return page_.delete(reason=comment)

        if param.parallel:
            return ut.parallelize(
                delete_single_page,
                pages,
                comment=param.comment,
                flush_at_end=param.debug,
            )
        else:
            return [delete_single_page(page, param.comment) for page in pages]

    class CreatePagePackageParam(OswBaseModel):
        """Parameter object for create_page_package method."""

        config: package.PagePackageConfig
        """Configuration object for the page package."""
        dump_config: Optional["WtPage.PageDumpConfig"] = None
        """Configuration object for the page dump"""
        debug: Optional[bool] = True
        """If True, debug messages will be printed."""
        parallel: Optional[bool] = None
        """If true, processes the pages in parallel."""

        class Config:
            arbitrary_types_allowed = True

    def create_page_package(self, param: CreatePagePackageParam):
        """Create a page package, which is a locally stored collection of wiki pages
        and their slots, based on a configuration object.

        Parameters:
        -----------
        param:  CreatePagePackageParam
        """

        debug = param.debug
        dump_config = param.dump_config
        config = param.config

        # Clear the content directory
        try:
            if debug:
                print(f"Delete dir '{config.content_path}'")
            if os.path.exists(config.content_path):
                shutil.rmtree(config.content_path)
        except OSError as e:
            if debug:
                print("Error: %s - %s." % (e.filename, e.strerror))
        # Create a dump config
        if dump_config is None:
            dump_config = WtPage.PageDumpConfig(
                target_dir=config.content_path,
                skip_slot_suffix_for_main=config.skip_slot_suffix_for_main,
            )
        else:
            dump_config.target_dir = config.content_path
            dump_config.skip_slot_suffix_for_main = config.skip_slot_suffix_for_main

        bundle: package.PagePackageBundle = config.bundle
        added_titles = []  # keep track of added pages, prevent duplicates

        if config.name not in bundle.packages:
            print(f"Error: package {config.name} does not exist in bundle")
            return
        if not bundle.packages[config.name].pages:
            bundle.packages[config.name].pages = []
        # remove duplicates while keeping order
        added_titles = list(dict.fromkeys(config.titles))

        pages = self.get_page(
            WtSite.GetPageParam(titles=added_titles, parallel=param.parallel)
        ).pages
        added_file_titles = []
        page_dumps = {}
        page_files = {}

        for page in pages:
            # Appends an item of type: package.PagePackagePage:
            page_dumps[page.title] = page.dump(dump_config)
            if config.include_files:
                referenced_file_pages = page.find_file_page_refs_in_slots()
                if debug and len(referenced_file_pages) > 0:
                    print(
                        f"File pages referenced in {page.title}: {referenced_file_pages}"
                    )
                if param.config.ignore_titles is not None:
                    included_file_pages = list(
                        set(referenced_file_pages) - set(param.config.ignore_titles)
                    )
                    # print ignored files
                    ignored_files_pages = list(
                        set(referenced_file_pages) - set(included_file_pages)
                    )
                    if debug and len(ignored_files_pages) > 0:
                        print(f"Ignored: {ignored_files_pages}")
                    referenced_file_pages = included_file_pages
                # find those files that are not already in the package
                page_files[page.title] = list(
                    set(referenced_file_pages) - set(added_file_titles)
                )
                added_file_titles = list(set(added_file_titles + referenced_file_pages))

        file_dumps = {}
        if len(added_file_titles) > 0:
            file_pages = self.get_page(
                WtSite.GetPageParam(titles=added_file_titles, parallel=param.parallel)
            ).pages
            for file_page in file_pages:
                if not file_page.exists:
                    continue
                # ToDo: Make this parallel as well
                file_dumps[file_page.title] = file_page.dump(dump_config)

        # bring it all together
        for page_title in added_titles:
            if page_title in page_dumps:
                bundle.packages[config.name].pages.append(page_dumps[page_title])
            if page_title in page_files:
                for file_title in page_files[page_title]:
                    if file_title in file_dumps:
                        bundle.packages[config.name].pages.append(
                            file_dumps[file_title]
                        )

        content = bundle.json(exclude_none=True, indent=4, ensure_ascii=False)
        # This will create the JSON (e.g., package.json) with the PagePackageConfig,
        #  which contains the PagePackageBundle
        file_name = f"{config.config_path}"
        with open(file_name, "w", encoding="utf-8") as f:
            f.write(content)

    class ReadPagePackageParam(OswBaseModel):
        """Parameter type of read_page_package."""

        storage_path: Union[str, Path]
        """The path to the directory where the page package is stored."""
        packages_info_file_name: Optional[Union[str, Path]] = None
        """The name of the file that contains the page package information. If not
            specified, the default value 'packages.json' is used."""
        selected_slots: Optional[List[str]] = None
        """A list of slots that should be read. If None, all slots are read."""
        debug: Optional[bool] = False
        """If True, debug information is printed to the console."""

    class ReadPagePackageResult(OswBaseModel):
        """Return type of read_page_package."""

        pages: List["WtPage"]
        """A list of WtPage objects."""

        class Config:
            arbitrary_types_allowed = True

    def read_page_package(self, param: ReadPagePackageParam) -> ReadPagePackageResult:
        """Read a page package, which is a locally stored collection of wiki pages and
        their slots' content.

        Parameters
        ----------
        param: ReadPagePackageParam

        Returns
        -------
        result: ReadPagePackageResult
        """
        # map params
        storage_path = param.storage_path
        packages_info_file_name = param.packages_info_file_name
        selected_slots = param.selected_slots
        debug = param.debug
        # Test arguments / set default value
        if not os.path.exists(storage_path):
            raise FileNotFoundError(f"Storage path '{storage_path}' does not exist.")
        if not isinstance(storage_path, Path):
            storage_path = Path(storage_path)
        if packages_info_file_name is None:
            packages_info_file_name = "packages.json"
        # Get top level content of storage path
        top_level_content = ut.list_files_and_directories(
            search_path=storage_path, recursive=False
        )
        # Try to find packages info file in storage path
        pi_fp = os.path.join(storage_path, packages_info_file_name)
        # Check if packages info file exists in storage path, otherwise use fallback
        #  option or raise error
        if os.path.exists(pi_fp) and os.path.isfile(pi_fp):
            if debug:
                print(f"Found packages info file at '{pi_fp}'.")
        else:
            if debug:
                print(
                    f"Did not find packages info file at '{pi_fp}'. Trying default "
                    f"'packages.json'."
                )
            top_level_json_files = [
                f
                for f in top_level_content["files"]
                if f.is_file() and str(f).endswith(".json")
            ]
            json_in_top_level = ut.file_in_paths(
                paths=top_level_json_files, file_name="packages.json"
            )
            if json_in_top_level["found"]:
                pi_fp = json_in_top_level["file path"]
            elif len(top_level_json_files) > 0:
                if debug:
                    print(f"Found JSON files: {top_level_json_files}. Using first one.")
                pi_fp = top_level_json_files[0]
            else:
                raise FileNotFoundError(
                    f"Error: No JSON files found in '{storage_path}'."
                )
        # Read packages info file
        with open(pi_fp, "r", encoding="utf-8") as f:
            packages_json = json.load(f)
        # Assume that the pages files are located in the subdir
        storage_path_content = ut.list_files_and_directories(
            search_path=storage_path, recursive=True
        )
        sub_dirs = top_level_content["directories"]
        if len(top_level_content["directories"]) == 0:
            # No subdirectories found, assume that the pages files are located in the
            #  top level
            sub_dirs = [storage_path]

        def get_slot_content(
            parent_dir: List[Union[str, Path]],
            url_path: str,
            files_in_storage_path: List[Path],
        ) -> Union[str, Dict]:
            for pdir in parent_dir:
                slot_path = storage_path / pdir / url_path
                if slot_path in files_in_storage_path:
                    with open(slot_path, "r", encoding="utf-8") as f:
                        file_content = f.read()
                    # Makes sure not to open an empty file with json
                    if len(file_content) > 0:
                        if url_path.endswith(".json"):
                            with open(slot_path, "r", encoding="utf-8") as f:
                                slot_data = json.load(f)
                            return slot_data
                        elif url_path.endswith(".wikitext"):
                            slot_data = file_content
                            return slot_data

        # Create WtPage objects
        pages = []
        for _package_name, package_dict in packages_json["packages"].items():
            for page in package_dict["pages"]:
                namespace = page["namespace"].split("_")[-1].capitalize()
                name = page["name"]
                # Create the WtPage object
                page_obj = WtPage(wtSite=self, title=f"{namespace}:{name}")
                if "main" in selected_slots:
                    # Main slot is special
                    slot_content = get_slot_content(
                        parent_dir=sub_dirs,
                        url_path=page["urlPath"],
                        files_in_storage_path=storage_path_content["files"],
                    )
                    if slot_content is not None:
                        page_obj.set_slot_content(
                            slot_key="main",
                            content=slot_content,
                        )
                if selected_slots is None:
                    _selected_slots = page["slots"]
                else:
                    _selected_slots = {
                        slot_name: slot_dict
                        for slot_name, slot_dict in page["slots"].items()
                        if slot_name in selected_slots
                    }
                for slot_name, slot_dict in _selected_slots.items():
                    slot_content = get_slot_content(
                        parent_dir=sub_dirs,
                        url_path=slot_dict["urlPath"],
                        files_in_storage_path=storage_path_content["files"],
                    )
                    if slot_content is not None:
                        page_obj.set_slot_content(
                            slot_key=slot_name,
                            content=slot_content,
                        )
                pages.append(page_obj)
        return WtSite.ReadPagePackageResult(pages=pages)

    class UploadPagePackageParam(OswBaseModel):
        """Parameter class for upload_page_package method."""

        storage_path: Optional[Union[str, Path]] = None
        """The path to the storage directory.
        If 'storage_path' is not given, 'pages' must be given."""
        pages: Optional[List["WtPage"]] = None
        """A list of WtPage objects.
        If 'pages' is not given, 'storage_path' must be given."""
        debug: Optional[bool] = False
        """If True, prints debug information."""

        class Config:
            arbitrary_types_allowed = True

    def upload_page_package(self, param: UploadPagePackageParam):
        """Uploads a page package to the wiki defined by a list of WtPage objects or
        a storage path.

        Parameters
        ----------
        param : UploadPagePackageParam

        """
        storage_path = param.storage_path
        pages = param.pages
        debug = param.debug

        if storage_path and pages is None:
            raise ValueError(
                "Error: If 'storage_path' is not given, 'pages' must be given."
            )
        if pages is None:
            pages = self.read_page_package(
                WtSite.ReadPagePackageParam(storage_path=storage_path, debug=debug)
            ).pages
        for page in pages:
            page.edit()

    def get_file_pages(self, limit: int = 1000000) -> List[str]:
        """Get all file pages in the wiki"""
        full_page_titles = wt.prefix_search(
            site=self._site,
            text=wt.SearchParam(query="File:", debug=False, limit=limit),
        )
        return full_page_titles

    def get_file_info_and_usage(
        self,
        page_titles: Union[str, List[str], SearchParam],
    ) -> list:
        """Get the file info and usage for one or more file pages.

        Parameters
        ----------
        page_titles:
            One or more full page titles of file pages or wiki_tools.SearchParam.

        Returns
        -------
        result:
            Dictionary with page titles as keys and nested dictionary with keys 'info'
            and 'usage'.
        """
        if isinstance(page_titles, str):
            title = wt.SearchParam(query=[page_titles], debug=False, parallel=True)
        elif isinstance(page_titles, list):
            title = wt.SearchParam(query=page_titles, debug=False, parallel=True)
        else:  # SearchParam
            title = page_titles
        return wt.get_file_info_and_usage(site=self._site, title=title)

    def get_prefix_dict(self):
        """Returns a dictionary with the custom prefixes used in osl"""
        prefix_dict = {
            "Property": "https://{domain}/id/Property-3A",
            "File": "https://{domain}/id/File-3A",
            "Category": "https://{domain}/id/Category-3A",
            "Item": "https://{domain}/id/Item-3A",
            "wiki": "https://{domain}/id/",
        }
        for prefix in prefix_dict:
            prefix_dict[prefix] = prefix_dict[prefix].replace(
                "{domain}", self._site.host
            )
        return prefix_dict

    def get_jsonld_context_prefixes(self):
        """Returns a dictionary with the custom prefixes used in osl for jsonld context"""
        context = {}
        for prefix, iri in self.get_prefix_dict().items():
            context[prefix] = {
                "@id": iri,
                "@prefix": True,
            }
        return context

    class JsonLdContextLoaderParams(OswBaseModel):
        prefer_external_vocal: Optional[bool] = True
        """Whether to prefer external vocabularies (e.g. skos, schema, etc.)
        over the local properties (Namespace 'Property:')"""

    def _replace_jsonld_context_mapping(
        self, context: Union[str, list, dict], config: JsonLdContextLoaderParams
    ):
        """
        interate over a jsonld context
        set <prefix> to the value of <prefix>* if not set yet
        handle string, list and dict values
        handle mappings direct to iri as well as
        {"@id": "http://example.org/property", @type": "@id"} and scoped contexts
        """
        if isinstance(context, str):
            return context
        if isinstance(context, list):
            return [self._replace_jsonld_context_mapping(e, config) for e in context]
        if isinstance(context, dict):
            context_iter = context.copy()
            for key in context_iter:
                value = context[key]
                if key == "wiki":
                    context[key] = f"https://{self._site.host}/id/"
                    # print(f"apply https://{self._site.host}/id/ to {key}")
                if isinstance(value, str):
                    base_key = key.split("*")[0]
                    if base_key not in context:
                        context[base_key] = value
                        # print(f"apply {key} to {base_key}")
                    if config.prefer_external_vocal is False:
                        base_mapping = context[base_key]
                        if isinstance(base_mapping, dict):
                            base_mapping = base_mapping["@id"]
                        mapping = value
                        if mapping.startswith(
                            "Property:"
                        ) and not base_mapping.startswith("Property:"):
                            context[base_key] = value
                            # print(f"apply {key} to {base_key}")
                elif isinstance(value, list):
                    context[key] = [
                        self._replace_jsonld_context_mapping(e, config) for e in value
                    ]
                elif isinstance(value, dict):
                    if "@id" in value:
                        base_key = key.split("*")[0]
                        if base_key not in context:
                            context[base_key] = value
                            # print(f"apply {key} to {base_key}")
                        if config.prefer_external_vocal is False:
                            base_mapping = context[base_key]
                            if isinstance(base_mapping, dict):
                                base_mapping = base_mapping["@id"]
                            mapping = value["@id"]
                            if mapping.startswith(
                                "Property:"
                            ) and not base_mapping.startswith("Property:"):
                                context[base_key] = value
                                # print(f"apply {key} to {base_key}")
                    elif "@context" in value:
                        context[key] = self._replace_jsonld_context_mapping(
                            value["@context"], config
                        )
            return context

    def get_jsonld_context_loader(self, params: JsonLdContextLoaderParams = None):
        if params is None:
            params = self.JsonLdContextLoaderParams()

        """to overwrite the default jsonld document loader to load
        relative context from the osl"""

        def loader(url, options=None):
            if options is None:
                options = {}
            # print("Requesting", url)
            if "/wiki/" in url:
                title = url.split("/wiki/")[-1].split("?")[0]
                page = self.get_page(WtSite.GetPageParam(titles=[title])).pages[0]
                if "JsonSchema:" in title:
                    schema = page.get_slot_content("main")
                else:
                    schema = page.get_slot_content("jsonschema")
                if isinstance(schema, str):
                    schema = json.loads(schema)
                schema["@context"] = self._replace_jsonld_context_mapping(
                    schema["@context"], params
                )

                doc = {
                    "contentType": "application/json",
                    "contextUrl": None,
                    "documentUrl": url,
                    "document": schema,
                }
                # print("Loaded", doc)
                return doc

            else:
                requests_loader = (
                    pyld.documentloader.requests.requests_document_loader()
                )
                return requests_loader(url, options)

        return loader


class WtPage:
    """A wrapper class of mwclient.page, mainly to provide multi-slot page handling"""

    def __init__(self, wtSite: WtSite = None, title: str = None, do_init: bool = True):
        """Creates a new WtPage object for the given title and loads the page from the
        site if the page already exists.

        Parameters
        ----------
        wtSite
            The instance site object, by default None
        title
            the page title, by default None
        do_init
            whether to initialize the page, by default True
        """
        self.wtSite = wtSite
        self.title = title

        self._original_content = ""
        self._content = ""
        self.changed: bool = False
        self._dict = []  # todo: named dict but is of type list
        self._slots: Dict[str, Union[str, dict]] = {"main": ""}
        self._slots_changed: Dict[str, bool] = {"main": False}
        self._content_model: Dict[str, str] = {"main": "wikitext"}

        if do_init:
            self.init()

    def init(self):
        """Initializes the page by loading the content and meta data from the site"""
        # inits the mwclient object - triggers an API call
        self._page = self.wtSite.mw_site.pages[self.title]
        self.exists = self._page.exists

        if self.exists:
            self._original_content = self._page.text()
            self._content = self._original_content
            # multi content revisions
            # second API call - ToDo: combine / replace with first call
            rev = self.wtSite.mw_site.api(
                "query",
                prop="revisions",
                titles=self.title,
                rvprop="ids|timestamp|flags|comment|user|content|contentmodel|roles|"
                "slotsize|slotsha1",
                rvslots="*",
                rvlimit="1",
                format="json",
            )
            for page_id in rev["query"]["pages"]:
                page = rev["query"]["pages"][page_id]
                if page["title"] == self.title:
                    for revision in page["revisions"]:
                        self._current_revision = revision
                        if "slots" in revision:
                            for slot_key in revision.get("slots", {}):
                                self._slots[slot_key] = revision["slots"][slot_key]["*"]
                                self._content_model[slot_key] = revision["slots"][
                                    slot_key
                                ]["contentmodel"]
                                self._slots_changed[slot_key] = False
                                # self._slots_sha1[slot_key] = \
                                #     revision["slots"][slot_key]["*"]
                                if self._content_model[slot_key] == "json":
                                    self._slots[slot_key] = json.loads(
                                        self._slots[slot_key]
                                    )
                        else:  # legacy MW instances < 1.35
                            self._slots["main"] = revision["*"]
                            self._content_model["main"] = "wikitext"
                            self._slots_changed["main"] = False
                            # self._slots_sha1["main"] = revision["sha1"]
                    # todo: set content for slots not in revision["slots"] (use
                    #  SLOTS) --> create empty slots

    def parse_main_slot(self):
        """Parses the main slot content of the page
        Requires wikitext dependencies installed with 'pip install osw[wikitext]'
        """
        self._dict = wt.create_flat_content_structure_from_wikitext(
            self._content, array_mode="only_multiple"
        )

    def create_slot(self, slot_key, content_model):
        """Creates a new slot for the page. Availables Keys and content models are defined in
        SLOTS.

        Parameters
        ----------
        slot_key
            The key of the slot (e.g. 'header', 'footer', 'jsondata', etc.)
        content_model
            The content model of the slot (wikitext, json, etc.)
        """
        self._slots[slot_key] = {}
        # To avoid TypeError: argument of type 'NoneType' is not iterable in
        #  set_slot_content()
        self._slots_changed[slot_key] = False
        self._content_model[slot_key] = content_model

    @deprecated("Use get_slot_content('main') instead")
    def get_content(self):
        """Get the content of the page (slot: main). Should be replaced by
        get_slot_content('main')
        Note: The content is parsed at page initialization with
        wt.create_flat_content_structure_from_wikitext()

        Returns
        -------
            _description_
        """
        return self._content

    def get_slot_content(self, slot_key, clone: bool = True):
        """Get the content of a slot

        Parameters
        ----------
        slot_key
            The key of the slot.
        clone
            Whether to return a pointer to the slot of the page or a clone (copy).

        Returns
        -------
            the content of the slot
        """
        if slot_key not in self._slots:
            return None
        if clone:
            return deepcopy(self._slots[slot_key])
        return self._slots[slot_key]

    def get_parsed_slot_content(self, slot_key: str) -> List:
        """Gets the parsed content of a slot by calling
        wt.create_flat_content_structure_from_wikitext()
        Parsed content can be set with set_parsed_slot_content().

        Parameters
        ----------
        slot_key
            the key of the slot

        Returns
        -------
            the parsed content of the slot
        """
        content = self.get_slot_content(slot_key)
        content = wt.create_flat_content_structure_from_wikitext(content)
        return content

    def get_slot_content_model(self, slot_key):
        """Get the content model of a slot

        Parameters
        ----------
        slot_key
            The key of the slot.

        Returns
        -------
            the content model of the slot
        """
        if slot_key not in self._slots:
            return None
        return self._content_model[slot_key]

    @deprecated("Use set_slot_content('main', content) instead")
    def set_content(self, content):
        """Sets the content of the page (slot: main). Should be replaced by
        set_slot_content('main', content)

        Parameters
        ----------
        content
            The new content of the page
        """
        self._content = content
        self.changed = True

    def set_slot_content(self, slot_key, content):
        """Sets the content of a slot

        Parameters
        ----------
        slot_key
            The key of the slot
        content
            the new content of the slot

        Raises
        ------
        ValueError
            if the slot_key is not defined in SLOTS
        """
        if slot_key not in self._slots:
            slot_dict = SLOTS.get(slot_key, None)
            if slot_dict is None:
                raise ValueError(
                    f"Error: Slot '{slot_key}' not defined in 'SLOTS'."
                    f"Available slots: {list(SLOTS.keys())}"
                )
            content_model = slot_dict["content_model"]
            self.create_slot(slot_key, content_model)
        if self._content_model[slot_key] == "json":
            new_content_str = json.dumps(content, ensure_ascii=False)
            original_content_str = json.dumps(self._slots[slot_key], ensure_ascii=False)
            if new_content_str != original_content_str:
                self._slots_changed[slot_key] = True
        else:
            if content != self._slots[slot_key]:
                self._slots_changed[slot_key] = True
        self._slots[slot_key] = content

    def set_parsed_slot_content(self, slot_key: str, content: List):
        """Sets the parsed content of a slot by calling
        wt.get_wikitext_from_flat_content_structure().
        Parsed content can be retrieved with get_parsed_slot_content()
        or wt.create_flat_content_structure_from_wikitext().

        Parameters
        ----------
        slot_key
            the key of the slot
        content
            the new parsed content of the slot
        """
        content = wt.get_wikitext_from_flat_content_structure(content)
        self.set_slot_content(slot_key, content)

    def get_url(self) -> str:
        """Get the URL of the page

        Returns
        -------
            the URL of the page
        """
        return "https://" + self.wtSite.mw_site.host + "/wiki/" + self.title

    def is_file_page(self) -> bool:
        """Checks if this page is a file page (containing an image, pdf, etc.)

        Returns
        -------
            true if this page is a file page, else false
        """
        return self.title.startswith("File:")

    @deprecated("No longer supported")
    def append_template(self, template_name: str = None, template_params: dict = None):
        """Appends a wiki template to the parsed main slot content.
        Please note that the main slots is parsed at page initialization
        with wt.create_flat_content_structure_from_wikitext().
        Remember to call update_content() afterwards

        Parameters
        ----------
        template_name
            the name of the template, by default None
        template_params
            the parameters of the template, by default None

        Returns
        -------
            the WtPage object
        """
        self._dict.append({template_name: template_params})
        return self

    @deprecated("No longer supported")
    def append_text(self, text):
        """Appends text to the parsed main slot content.
        Please note that the main slots is parsed at page initialization
        with wt.create_flat_content_structure_from_wikitext().
        Remember to call update_content() afterwards

        Parameters
        ----------
        text:
            the text to append

        Returns
        -------
            the WtPage object
        """
        self._dict.append(text)
        return self

    @deprecated("No longer supported")
    def get_value(self, jsonpath):
        """Resolves a JSONPath expression in the parsed main slot content.
        Please note that the main slots is parsed at page initialization
        with wt.create_flat_content_structure_from_wikitext().

        Parameters
        ----------
        jsonpath
            The JSONPath expression

        Returns
        -------
            The query result
        """
        jsonpath_expr = parse(jsonpath)
        res = []
        d = dict(
            zip(range(len(self._dict)), self._dict)
        )  # convert list to dict with index
        for match in jsonpath_expr.find(d):
            res.append(match.value)
        return res

    @deprecated("No longer supported")
    def update_dict(self, combined: dict, update: dict) -> None:
        for k, v in update.items():
            if isinstance(v, dict):
                # todo: fix reference for combine_into
                wt.combine_into(v, combined.setdefault(k, {}))
            else:
                combined[k] = v

    @deprecated("No longer supported for replace=False")
    def set_value(self, jsonpath_match, value, replace=False):
        """Sets the value of a JSONPath expression in the parsed main slot content.
        Please note that the main slots is parsed at page initialization
        with wt.create_flat_content_structure_from_wikitext().
        Remember to call update_content() afterwards

        Parameters
        ----------
        jsonpath_match
            The JSONPath expression
        value
            The value to set
        replace
            Whether to replace the value, by default False

        Returns
        -------
            The WtPage object
        """
        jsonpath_expr = parse(jsonpath_match)
        d = dict(
            zip(range(len(self._dict)), self._dict)
        )  # convert list to dict with index
        # if create: jsonpath_expr.update_or_create(d, value)
        # else: jsonpath_expr.update(d, value)
        matches = jsonpath_expr.find(d)
        for match in matches:
            print(match.full_path)
            # pprint(value)
            if not replace:
                WtPage.update_dict(match.value, value)
                value = match.value
            # pprint(value)
            match.full_path.update_or_create(d, value)
        self._dict = list(d.values())  # convert dict with index to list
        return self

    @deprecated("Use set_parsed_slot_content('main', content) instead")
    def update_content(self):
        """Updates the content of the page with the parsed content of the main slots
        by calling wt.get_wikitext_from_flat_content_structure().

        Returns
        -------
            the WtPage object
        """
        self._content = wt.get_wikitext_from_flat_content_structure(self._dict)
        self.changed = self._original_content != self._content
        return self

    def edit(self, comment: str = None, mode="action-multislot"):
        """Creates / updates the content of all page slots in the wiki site

        Parameters
        ----------
        comment:
            (optional) edit comment for the page history, by default None
        mode:
            (optional) single API call ('action-multislot') or multiple (
            'action-singleslot'), by default 'action-multislot' (faster)
        """
        retry = 0
        max_retry = 5
        while retry < max_retry:
            try:
                return self._edit(comment, mode)
            except Exception as e:
                if retry < max_retry:
                    retry += 1
                    print(f"Page edit failed: {e}. Retry ({retry}/{max_retry})")
                    sleep(5)

    def _edit(self, comment: str = None, mode="action-multislot"):
        if not comment:
            comment = "[bot] update of page content"
        if self.changed:
            self._page.edit(self._content, comment)  # legacy mode
        if mode == "action-multislot":
            params = {}
            changed = False
            for slot_key in self._slots:
                if self._slots_changed[slot_key]:
                    changed = True
                    self._slots_changed[slot_key] = False
                    content = self._slots[slot_key]
                    if self._content_model[slot_key] == "json":
                        if not isinstance(content, str):
                            content = json.dumps(content, ensure_ascii=False)
                    params["slot_" + slot_key] = content
            if changed:
                self.changed = True
                self.wtSite.mw_site.api(
                    "editslots",
                    token=self.wtSite.mw_site.get_token("csrf"),
                    title=self.title,
                    summary=comment,
                    **params,
                )
                self.wtSite._clear_cookies()

        else:
            changed = False
            for slot_key in self._slots:
                if self._slots_changed[slot_key]:
                    changed = True
                    content = self._slots[slot_key]
                    if self._content_model[slot_key] == "json":
                        content = json.dumps(content, ensure_ascii=False)
                    self.wtSite.mw_site.api(
                        "editslot",
                        token=self.wtSite.mw_site.get_token("csrf"),
                        title=self.title,
                        slot=slot_key,
                        text=content,
                        summary=comment,
                    )
                    self._slots_changed[slot_key] = False
            if changed:
                self.changed = True

    def delete(self, comment: str = None):
        """Deletes the page from the site

        Parameters
        ----------
        comment
            The delete comment, by default None
        """
        self._page.delete(comment)

    def move(self, new_title: str, comment: str = None, redirect: bool = True):
        """Moves (=renames) the page to a new title

        Parameters
        ----------
        new_title
            the new title of the page
        comment
            the edit comment
        redirect
            whether to create a redirect from the old title to the new title
        """
        if new_title != self.title:
            print(f"move '{self.title}' to '{new_title}'")
            self._page.move(
                new_title=new_title, reason=comment, no_redirect=not redirect
            )
            self.title = new_title

    def get_last_changed_time(self):
        """Gets the timestamp of the last change of the page

        Returns
        -------
            a datetime string in ISO format
        """
        return datetime.fromisoformat(
            self._current_revision["timestamp"].replace("Z", "+00:00")
        )

    class CopyPageConfig(OswBaseModel):
        """Configuration to copy a page"""

        source_site: WtSite
        """The source site to copy the page from"""
        existing_page: str
        """The full page title of the page on the source site"""
        overwrite: Optional[bool] = False
        """If true, the page will be overwritten if it already exists on the target
        site"""
        comment: Optional[str] = None
        """Edit comment for the page history"""

        class Config:
            arbitrary_types_allowed = True

    class PageCopyResult(OswBaseModel):
        """Result of copying a page"""

        page: "WtPage"
        """The copied page"""
        target_altered: bool
        """True if the page at the target site was altered"""

        class Config:
            arbitrary_types_allowed = True

    def copy(self, config: CopyPageConfig) -> PageCopyResult:
        if config.comment is None:
            config.comment = f"[bot edit] Copied from {config.source_site.mw_site.host}"
        result = config.source_site.get_page_content([config.existing_page])
        for title, slot_contents in result.contents.items():
            self.title = title
            verb = "created"
            if self.exists:
                verb = "updated"
                if config.overwrite is False:
                    return WtPage.PageCopyResult(page=self, target_altered=False)
                changed_slots = []
                for slot in SLOTS:
                    if self.get_slot_content(slot) != slot_contents.get(slot, None):
                        changed_slots.append(slot)
                if len(changed_slots) == 0:
                    print(
                        f"Page '{self.title}' already has the same content. It will "
                        f"not be updated."
                    )
                    return WtPage.PageCopyResult(page=self, target_altered=False)
                else:
                    print(
                        f"Page '{self.title}' has different content in slots "
                        f"{changed_slots}."
                    )
            for slot in slot_contents.keys():
                self.create_slot(
                    slot_key=slot, content_model=SLOTS[slot]["content_model"]
                )
                self.set_slot_content(slot_key=slot, content=slot_contents[slot])
            self.edit(comment=config.comment)
            s2p = (
                f"Page {verb}: "
                f"'https://{self.wtSite.mw_site.host}/wiki/{self.title}'."
            )
            if verb == "updated":
                s2p = (
                    f"Page {verb}: "
                    f"'https://{self.wtSite.mw_site.host}/w/index.php?title"
                    f"={self.title}&action=history'."
                )
            print(s2p)
            return WtPage.PageCopyResult(page=self, target_altered=True)

    class PageDumpConfig(OswBaseModel):
        """Configuration to dump wiki pages to the file system"""

        target_dir: Union[str, Path]
        """Directory to dump all contents.
        Will be created automatically if not existing"""
        namespace_as_folder: Optional[bool] = True
        """Store page contents in subfolders named according to their namespaces"""
        skip_slot_suffix_for_main: Optional[bool] = False
        """If true, do not include 'main' in the generated content file.
        Useful for pages / wikis using only the main slot."""
        dump_empty_slots: Optional[bool] = False
        """If true, dump all configured slots even empty ones.
        Useful to create initial content in these slots."""
        page_name_as_filename: Optional[bool] = False
        """Use the (human readable) name of a page also for the file naming.
        Useful to identify dumped files manually. The mapping to the page title/id
        (in general a UUID) is ensured anyway through package.json meta data."""

        class Config:
            arbitrary_types_allowed = True  # necessary to allow e.g. np.array as type

    def dump(self, config: PageDumpConfig) -> package.PagePackagePage:
        """Dump this page to the file system

        Parameters
        ----------
        config
            see WtPage.PageDumpConfig

        Returns
        -------
            Metadata of the generated dump
        """
        page_name = self.title.split(":")[-1]
        if ":" in self.title:
            namespace = self.title.split(":")[0]
        else:
            namespace = "Main"

        namespace_const = "NS_" + namespace.upper()
        if namespace in package.NAMESPACE_TO_NAMESPACE_CONST_MAPPING:
            namespace_const = package.NAMESPACE_TO_NAMESPACE_CONST_MAPPING[namespace]

        package_page = package.PagePackagePage(
            name=page_name, namespace=namespace_const, slots={}
        )
        name_in_json_data = False
        if "jsondata" in self._slots and "name" in self._slots["jsondata"]:
            package_page.label = self._slots["jsondata"]["name"]
            name_in_json_data = True

        if config.page_name_as_filename and name_in_json_data:
            # Use name from jsondata as filename for the dump
            dump_name = self._slots["jsondata"]["name"]
        else:
            # Use page name as filename for the dump
            dump_name = page_name

        tar_dir = config.target_dir
        path_prefix = ""
        if config.namespace_as_folder:
            tar_dir = os.path.join(tar_dir, namespace)
            path_prefix = namespace + "/"
        # if not os.path.exists(dir):
        #    os.makedirs(dir)

        def save_to_file(file_path__, content__):
            if not os.path.exists(os.path.dirname(file_path__)):
                os.makedirs(os.path.dirname(file_path__))
            with open(os.path.join(file_path__), "w", encoding="utf-8") as f__:
                f__.write(content__)

        def dump_slot_content(slot_key_, content_type_, content_):
            if isinstance(content_, dict):
                content_ = json.dumps(content_, indent=4, ensure_ascii=False)
            if content_type_ == "Scribunto":
                content_type_ = "lua"
            if slot_key_ == "main" and config.skip_slot_suffix_for_main:
                file_name_ = f"{dump_name}.{content_type_}"
            else:
                file_name_ = f"{dump_name}.slot_{slot_key_}.{content_type_}"
            # handle subpages:
            file_path_ = os.path.join(tar_dir, *file_name_.split("/"))
            save_to_file(file_path_, content_)
            # generate url encoded path for the page package
            if slot_key_ == "main":
                package_page.urlPath = urllib.parse.quote(path_prefix + file_name_)
            else:
                package_page.slots[slot_key_] = package.PagePackagePageSlot(
                    urlPath=urllib.parse.quote(path_prefix + file_name_)
                )

        for slot_key in self._slots:
            content = self.get_slot_content(slot_key)
            content_type = self.get_slot_content_model(slot_key)
            dump_slot_content(slot_key, content_type, content)

        # If the slots are empty, we still want files to fill after dumping them
        if config.dump_empty_slots:
            for slot_key in [slot for slot in SLOTS.keys() if slot not in self._slots]:
                content = SLOTS[slot_key]["content_template"]
                content_type = SLOTS[slot_key]["content_model"]
                dump_slot_content(slot_key, content_type, content)

        if self.is_file_page():
            file = self.wtSite.mw_site.images[self.title.split(":")[-1]]
            file_name = f"{page_name}"
            file_path = os.path.join(tar_dir, *file_name.split("/"))  # handle subpages
            # The following will return KeyError "url" if the file is not found
            with open(file_path, "wb") as fd:
                file.download(fd)
            package_page.fileURLPath = path_prefix + file_name

        return package_page

    def get_file_info_and_usage(
        self, debug: bool = False
    ) -> Dict[str, Union[str, List[str]]]:
        """For file page only: Get the file info and usage for this file page

        Parameters
        ----------
        debug
            Whether to print debug information, by default False

        Returns
        -------
            Dictionary with page titles as keys and nested dictionary with keys 'info'
            and 'usage'.
        """
        return wt.get_file_info_and_usage(
            site=self.wtSite.mw_site,
            title=wt.SearchParam(query=self.title, debug=debug),
        )[0]

    def find_file_page_refs_in_slots(self, slots: List[str] = None) -> List[str]:
        """Find all file page references in the content of the given slots."""
        if slots is None:
            slots = list(self._slots.keys())
        file_page_refs = []
        for slot in slots:
            # get slot content
            content = self.get_slot_content(slot)
            if content is None:
                continue
            # find all file page references
            pattern = REGEX_PATTERN_LIB["File page strings from any text"]
            full_page_names = pattern.findall_by_group_key(
                str(content),  # make sure its a string not a dictionary
                "Full page name",
            )
            file_page_refs.extend(full_page_names)
            # Find all files in editor templates
            pattern = REGEX_PATTERN_LIB["File uuid in template"]
            res = pattern.finditer(str(content))
            # Iterate over all matches
            for match in res:
                ft = None
                # Check if the match has the groups "Editor" and "UUID" and UUID is a
                #  valid OSW ID
                try:
                    if "Editor" in match.groups and "UUID" in match.groups:
                        # construct a file page title
                        osw_id = get_osw_id(match.groups["UUID"])
                        if match.groups["Editor"] == "DrawIO":
                            ft = "File:" + osw_id + ".drawio.svg"
                        elif match.groups["Editor"] == "SvgEdit":
                            ft = "File:" + osw_id + ".svg"
                        elif match.groups["Editor"] == "Kekule":
                            ft = "File:" + osw_id + ".kekule.json"
                        elif match.groups["Editor"] == "Spreadsheet":
                            ft = "File:" + osw_id + ".luckysheet.json"
                        elif match.groups["Editor"] == "Wellplate":
                            ft = "File:" + osw_id + ".wellplate.svg"
                    if ft is not None:
                        file_page_refs.append(ft)
                except ValueError:
                    print("Warning: Error while parsing uuid in editor template")
        return list(set(file_page_refs))

    def purge(self):
        """Purge the page from the site cache.
        Triggers a rebuild / refresh of the page.
        This is useful if the page content is changed and the changes are not visible
        """
        self._page.purge()

    class ExportConfig(OswBaseModel):
        """Configuration to export a page to XML"""

        full_history: Optional[bool] = True
        """if true, export the full history of the page, else only the current revision"""
        include_templates: Optional[bool] = False
        """if true, export the templates used in the page"""

    class ExportResult(OswBaseModel):
        """Return type of export_xml"""

        xml: str
        """the XML string"""
        success: bool
        """if true, the export was successful, else false"""

    def export_xml(self, config: Optional[ExportConfig] = None) -> ExportResult:
        """Exports the page to XML

        Parameters
        ----------
        config
            see ExportConfig

        Returns
        -------
            ExportResult
        """
        if config is None:
            config = WtPage.ExportConfig()
        url = (
            self.wtSite.mw_site.scheme
            + "://"
            + self.wtSite.mw_site.host
            + self.wtSite.mw_site.path
            + "index.php?title=Special:Export/"
            + self.title
        )
        data = {
            "title": "Special:Export",
            "catname": "",
            "pages": self.title,
            "wpEditToken": self.wtSite.mw_site.get_token("csrf"),
            "wpDownload": "1",
        }
        if not config.full_history:
            data["curonly"] = "1"
        if config.include_templates:
            data["templates"] = "1"
        response = self.wtSite.mw_site.connection.post(url, data=data)
        if response.status_code != 200:
            return WtPage.ExportResult(success=False, xml="")
        else:
            return WtPage.ExportResult(success=True, xml=response.text)

    class ImportConfig(OswBaseModel):
        """Configuration to import a page from XML.
        see also https://www.mediawiki.org/wiki/Manual:Importing_XML_dumps"""

        xml: str
        """the XML string to import (see WtPage.export_xml)"""
        summary: str
        """the edit summary to use for the import"""
        source_domain: str
        """the domain of the instance from which the XML was exported, e.g. mywiki.com"""
        full_history: Optional[bool] = True
        """if true, import the full history of the page, else only the current revision"""
        include_templates: Optional[bool] = False
        """if true, import the templates used in the page if contained in the XML"""
        namespace_mapping: Optional[Dict[str, str]] = {
            "Main": 0,
            "File": 6,
            "Template": 10,
            "Category": 14,
            "Item": 7000,
        }
        """mapping of namespaces names to IDs in the target instance"""
        username_mapping: Optional[Dict[str, str]] = {}
        """mapping of usernames in the XML to usernames in the target instance"""

    class ImportResult(OswBaseModel):
        """Return type of import_xml"""

        success: bool
        """if true, the import was successful, else false"""
        imported_title: str
        imported_revisions: int
        error_msg: Optional[str] = None

    def import_xml(self, config: ImportConfig) -> ImportResult:
        """Imports the page from an XML export

        Parameters
        ----------
        config
            see ImportConfig

        Returns
        -------
            ExportResult
        """

        # Remove default namespace definition (see https://stackoverflow.com/questions/
        #  34009992/python-elementtree-default-namespace)
        config.xml = config.xml.replace(
            'xmlns="http://www.mediawiki.org', '_xmlns="http://www.mediawiki.org'
        )
        print(config.xml)
        tree = et.fromstring(config.xml)

        # Replace title and namespace with the requested ones
        tree.find(".//title").text = self.title.split(":")[1]
        tree.find(".//ns").text = str(
            config.namespace_mapping.get(self.title.split(":")[0], 0)
        )
        # Apply username mapping (user in the target system might have different names)
        for e in tree.findall(".//username"):
            e.text = config.username_mapping.get(e.text, e.text)

        config.xml = et.tostring(tree, encoding="unicode")
        # Restore default namespace definition
        config.xml = config.xml.replace(
            '_xmlns="http://www.mediawiki.org', 'xmlns="http://www.mediawiki.org'
        )

        api_url = (
            self.wtSite.mw_site.scheme
            + "://"
            + self.wtSite.mw_site.host
            + self.wtSite.mw_site.path
            + "api.php"
        )
        response = self.wtSite.mw_site.connection.post(
            url=api_url,
            data={
                "action": "import",
                "token": self.wtSite.mw_site.get_token("csrf"),
                "fullhistory": "1" if config.full_history else "0",
                "templates": "1" if config.include_templates else "0",
                "assignknownusers": "1",
                "interwikiprefix": config.source_domain,
                # "namespace": self.title.split(":")[0],
                "summary": config.summary,
                "format": "json",
            },
            files={
                "xml": (
                    "xml",
                    StringIO(config.xml),
                    "text/xml",
                )  # read config.xml as file
            },
        )

        json = response.json()
        if "error" in json:
            # print("Error: ", json)
            return WtPage.ImportResult(success=False, error_msg=json["error"]["info"])
        else:
            # print("Imported: ", json["import"][0]["title"], " with ",
            #       json["import"][0]["revisions"], " revisions")
            return WtPage.ImportResult(
                success=True,
                imported_title=json["import"][0]["title"],
                imported_revisions=json["import"][0]["revisions"],
            )


# Updating forwards refs in pydantic models
WtPage.PageCopyResult.update_forward_refs()
WtSite.CopyPagesParam.update_forward_refs()
WtSite.UploadPageParam.update_forward_refs()
WtSite.GetPageResult.update_forward_refs()
WtSite.CreatePagePackageParam.update_forward_refs()
WtSite.UploadPagePackageParam.update_forward_refs()
WtSite.ReadPagePackageResult.update_forward_refs()
WtSite.DeletePageParam.update_forward_refs()



import sys
from pathlib import Path

import pytest

import osw.model.entity as model
from osw.auth import CredentialManager
from osw.core import OSW, AddOverwriteClassOptions, OverwriteOptions
from osw.utils.wiki import get_full_title
from osw.wiki_tools import SearchParam
from osw.wtsite import WtSite

# run with: tox -e test -- --wiki_domain domain --wiki_username user --wiki_password pass


def test_store_and_load(wiki_domain, wiki_username, wiki_password):
    cm = CredentialManager()
    cm.add_credential(
        CredentialManager.UserPwdCredential(
            iri=wiki_domain, username=wiki_username, password=wiki_password
        )
    )
    wtsite = WtSite(WtSite.WtSiteConfig(iri=wiki_domain, cred_mngr=cm))
    osw = OSW(site=wtsite)
    # Check 1: Store an entity and download it again, delete afterward
    # Create an item with a label
    my_entity = model.Item(label=[model.Label(text="MyItem")])
    # Store the item in the OSW
    osw.store_entity(my_entity)
    # Load the item from the OSW
    my_entity2 = osw.load_entity("Item:" + OSW.get_osw_id(my_entity.uuid))
    # Check the stored item
    assert my_entity.label[0].text == my_entity2.label[0].text
    # Delete the item
    osw.delete_entity(my_entity)

    # Check 2: Store a more complex entity, create a local duplicate, with changed
    # properties, test the 'overwrite' param options, delete afterward
    # Make non-package scripts available for import
    cwd = Path(__file__).parent.absolute()
    tests_dir = cwd.parents[1] / "tests"
    sys.path.append(str(tests_dir))
    # Get required functions
    from tests.test_osl import (
        check_false,
        check_keep_existing,
        check_only_empty,
        check_replace_remote,
        check_true,
    )

    checks = [
        {  # Overwrite properties
            "overwrite": OverwriteOptions.true,
            "assert": check_true,
        },
        {  # Do not overwrite properties
            "overwrite": OverwriteOptions.false,
            "assert": check_false,
        },
        {  # Overwrite empty properties only
            "overwrite": OverwriteOptions.only_empty,
            "assert": check_only_empty,
        },
        {  # Replace the remote entity entirely
            "overwrite": AddOverwriteClassOptions.replace_remote,
            "assert": check_replace_remote,
        },
        {  # Keep the existing entity as is
            "overwrite": AddOverwriteClassOptions.keep_existing,
            "assert": check_keep_existing,
        },
    ]
    for check in checks:
        # Create a new item with some properties
        original_item = model.Item(
            label=[model.Label(text="My Item")],
            name="MyItem",
            iri="",  # Empty string property
            description=[],  # Empty list property
            query_label=None,  # Equal to a non-existing property
            image="File:OSWacacwdcawd.png",  # Property not in altered
        )
        print("Storing original entity...")
        osw.store_entity(original_item)
        # Alter some of the property values
        altered_props = {
            "label": [model.Label(text="My Item Duplicate")],
            "name": "MyItemDuplicate",
            "iri": "http://example.com/MyItemDuplicate",
            "description": [model.Label(text="This is a duplicate")],
            "query_label": "My Item Duplicate",
            "attachments": ["File:OSWacacwdcawd.pdf"],  # Property not in original
        }
        # Create a new item with the altered properties
        original_props = {
            key: value
            for key, value in original_item.dict().items()
            if key in ["uuid", "type"]
        }
        altered_item = model.Item(**{**original_props, **altered_props})
        # Update the item in the OSW
        print("Storing altered entity...")
        osw.store_entity(
            param=OSW.StoreEntityParam(
                entities=altered_item, overwrite=check["overwrite"]
            )
        )
        # Load the item from the OSW
        stored_item = osw.load_entity(
            entity_title=OSW.LoadEntityParam(
                titles="Item:" + OSW.get_osw_id(original_item.uuid), disable_cache=True
            )
        ).entities[0]
        # Check the stored item
        check["assert"](original_item, altered_item, stored_item)
        # Delete the item
        osw.delete_entity(original_item)


def test_query_instances(wiki_domain, wiki_username, wiki_password):
    """Store an entity, query instances of the category of the entity, make sure the
    new entity is contained in the list of returned instances, delete the entity."""
    cm = CredentialManager()
    cm.add_credential(
        CredentialManager.UserPwdCredential(
            iri=wiki_domain, username=wiki_username, password=wiki_password
        )
    )
    wtsite = WtSite(WtSite.WtSiteConfig(iri=wiki_domain, cred_mngr=cm))
    osw = OSW(site=wtsite)
    # Create an item with a label
    my_item = model.Item(label=[model.Label(text="MyItem")])
    fpt = get_full_title(my_item)
    # Store the item in the OSW
    osw.store_entity(my_item)
    # Query instances of the category of the entity
    instances = osw.query_instances(
        category=OSW.QueryInstancesParam(categories="Category:Item", limit=10000)
    )
    assert fpt in instances
    # Delete the item
    osw.delete_entity(my_item)


@pytest.mark.skip(reason="Temporarily disabled - Failing due to unresolved error")
def test_statement_creation(wiki_domain, wiki_username, wiki_password):
    cm = CredentialManager()
    cm.add_credential(
        CredentialManager.UserPwdCredential(
            iri=wiki_domain, username=wiki_username, password=wiki_password
        )
    )
    wtsite = WtSite(WtSite.WtSiteConfig(iri=wiki_domain, cred_mngr=cm))
    osw = OSW(site=wtsite)

    my_entity = model.Item(
        label=[model.Label(text="MyItem")],
        statements=[
            model.DataStatement(property="Property:TestProperty", value="TestValue")
        ],
    )

    osw.store_entity(my_entity)

    search_param = SearchParam(query="[[TestProperty::TestValue]]")
    full_page_titles = osw.site.semantic_search(search_param)
    assert f"Item:{OSW.get_osw_id(my_entity.uuid)}" in full_page_titles

    search_param = SearchParam(
        query="[[HasStatement.HasProperty::Property:TestProperty]]"
    )
    full_page_titles = osw.site.semantic_search(search_param)
    assert f"Item:{OSW.get_osw_id(my_entity.uuid)}" in full_page_titles

    osw.delete_entity(my_entity)


def test_characteristic_creation(wiki_domain, wiki_username, wiki_password):
    cm = CredentialManager()
    cm.add_credential(
        CredentialManager.UserPwdCredential(
            iri=wiki_domain, username=wiki_username, password=wiki_password
        )
    )
    wtsite = WtSite(WtSite.WtSiteConfig(iri=wiki_domain, cred_mngr=cm))
    osw = OSW(site=wtsite)

    osw.fetch_schema(
        OSW.FetchSchemaParam(
            schema_title=[
                "Category:OSWffe74f291d354037b318c422591c5023",  # CharacteristicType
                "Category:Property",
            ],
            mode="replace",
        )
    )

    # my_property = model.Property

    # Create a characteristic as instance of CharacteristicType
    my_characteristic = model.CharacteristicType(
        uuid="efad8086-4a76-47ca-b278-f5f944e5754b",
        name="TestCharacteristic",
        label=[model.Label(text="Test Characteristic")],
        properties=[
            model.PrimitiveProperty(
                uuid="766e7171-a183-4f9c-a9af-28cfd27fb1d9",
                name="test_property",
                type="string",
                property_type="SimpleProperty",
            ),
            model.PrimitiveProperty(
                uuid="766e7171-a183-4f9c-a9af-28cfd27fb1d1",
                name="test_property2",
                rdf_property="Property:TestPropertyWithSchema",
                property_type="SimpleProperty",
            ),
        ],
    )

    # store it as instance, which generates the schema
    # note: namespace and meta_category should be detected automatically in the future
    osw.store_entity(
        OSW.StoreEntityParam(
            entities=[my_characteristic],
            namespace="Category",
            meta_category_title="Category:OSWffe74f291d354037b318c422591c5023",
            overwrite=OverwriteOptions.true,
        )
    )

    pages = osw.site.get_page(
        WtSite.GetPageParam(titles=["Property:TestPropertyWithSchema"])
    ).pages
    pp = pages[0]
    # schema: dict = pp.get_slot_content("jsonschema")
    new_schema = {"type": "number", "description": "Imported from property schema"}
    pp.set_slot_content("jsonschema", new_schema)
    pp.edit()

    # load the characteristic as category
    osw.fetch_schema(
        OSW.FetchSchemaParam(
            schema_title=[
                "Category:OSWefad80864a7647cab278f5f944e5754b"  # TestCharacteristic
            ],
            mode="append",
        )
    )

    # create an instance of the characteristic
    t = model.TestCharacteristic(test_property="Test", test_property2=1)
    assert t.test_property == "Test"
    assert t.test_property2 == 1

    # cleanup (disabled for paralle test matrix)
    # my_characteristic.meta = model.Meta(
    #     wiki_page=model.WikiPage(namespace="Category")
    # )  # namespace detection fails otherwise
    # osw.delete_entity(my_characteristic)
    # pp.delete()



import os
from pprint import pprint
from typing import List, Optional

from jsonpath_ng.ext import parse

import osw.model.entity as model
from osw.express import OswExpress

# Create/update the password file under examples/accounts.pwd.yaml
pwd_file_path = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "accounts.pwd.yaml"
)
# pwd_file_path = "./accounts.pwd.yaml"
osw_obj = OswExpress(
    domain="wiki-dev.open-semantic-lab.org", cred_filepath=pwd_file_path
)

entities_dict = {}  # Helper dict for jsonpath queries


# Target class
class SomeItem(model.Item):
    type: List[str] = ["Category:Item"]
    custom_attribute: Optional[str]


# Helper class
class CustomItem(SomeItem):
    internal_id: str


# Construct dict from semantic query
# entities = osw_obj.site.semantic_search("[[HasType::Category:Item]]")
# for entity in entities:
#   entities_dict[entity] = {
#   'jsondata': osw_obj.load_entity(entity).json(exclude_none=True)
#   }

# Manually create test data
e1 = CustomItem(label=[model.Label(text="Item1")], internal_id="001")
e2 = CustomItem(
    label=[model.Label(text="Item2")], internal_id="002", custom_attribute="001"
)
entities = [e1, e2]

print("\nInitial data with internal IDs:")
for e in entities:
    print(e.cast(SomeItem).json(exclude_none=True))  # Casting strips 'internal_id'
    entities_dict[str(e.uuid)] = {}
    entities_dict[str(e.uuid)]["jsondata"] = e.dict(
        exclude_none=True
    )  # Query does not work without intermediate attribute 'jsondata'

print("\nHelper dict")
pprint(entities_dict)

print("\nProcessed data with global IDs:")
for e in entities:
    jsonpath_internal_id_match = parse(f'*[?internal_id = "{e.custom_attribute}"]')
    matches = jsonpath_internal_id_match.find(entities_dict)
    if len(matches) > 0:
        e.custom_attribute = str(
            matches[0].value["uuid"]
        )  # Replace internal id with uuid
    if len(matches) > 1:
        print(
            "Warning: Multiple matches for "
            + f'*[?internal_id = "{e.custom_attribute}"]'
        )
    print(e.cast(SomeItem).json(exclude_none=True))  # Casting strips 'internal_id'



from osw.express import OswExpress

# Create an OswExpress object
domain = "wiki-dev.open-semantic-lab.org"
osw = OswExpress(domain=domain)
instances = osw.site.semantic_search(
    "[[Category:OSW44deaa5b806d41a2a88594f562b110e9]]"  # Persons ins OSW
    "[[HasOu::Item:OSWb961cad90656513babc4e928472538c4]]"  # that are member of the OU
)
print(instances)



from __future__ import annotations

import importlib
import json
import os
import platform
import re
import sys
from copy import deepcopy
from enum import Enum
from typing import Any, Dict, List, Optional, Type, Union
from uuid import UUID, uuid4
from warnings import warn

import rdflib
from jsonpath_ng.ext import parse
from mwclient.client import Site
from pydantic.v1 import BaseModel, PrivateAttr, create_model, validator
from pyld import jsonld

import osw.model.entity as model
from osw.model.static import OswBaseModel
from osw.utils.templates import (
    compile_handlebars_template,
    eval_compiled_handlebars_template,
)
from osw.utils.util import parallelize
from osw.utils.wiki import (
    get_full_title,
    get_namespace,
    get_title,
    get_uuid,
    is_empty,
    namespace_from_full_title,
    remove_empty,
    title_from_full_title,
)
from osw.wiki_tools import SearchParam
from osw.wtsite import WtPage, WtSite


# Reusable type definitions
class OverwriteOptions(Enum):
    """Options for overwriting properties"""

    true = True
    """Always overwrite a property"""
    false = False
    """Never overwrite a property"""
    only_empty = "only empty"
    """Only overwrite if the property is empty"""
    # todo: implement "merge",
    # todo: implement "append",
    #   Don't replace the properties but for properties of type array or dict, append
    #   the values of the local entity to the remote entity, make sure  # to not
    #   append duplicates
    # todo: implement "only older",
    # todo: implement read out from the version history of the page


class AddOverwriteClassOptions(Enum):
    replace_remote = "replace remote"
    """Replace the entity with the new one, removes all properties that are not
    present in the local entity"""
    keep_existing = "keep existing"
    """Keep the entity, does not add or remove any properties, if the page exists, the
    entity is not stored"""
    none = None
    """Not an option to choose from, will be replaced by the default remote properties
    """


OVERWRITE_CLASS_OPTIONS = Union[OverwriteOptions, AddOverwriteClassOptions]


class OSW(BaseModel):
    """Bundles core functionalities of OpenSemanticWorld (OSW)"""

    uuid: str = "2ea5b605-c91f-4e5a-9559-3dff79fdd4a5"
    _protected_keywords = (
        "_osl_template",
        "_osl_footer",
    )  # private properties included in model export

    class Config:
        arbitrary_types_allowed = True  # necessary to allow e.g. np.array as type

    site: WtSite

    @property
    def mw_site(self) -> Site:
        """Returns the mwclient Site object of the OSW instance."""
        return self.site.mw_site

    @staticmethod
    def get_osw_id(uuid: uuid) -> str:
        """Generates a OSW-ID based on the given uuid by prefixing "OSW" and removing
        all '-' from the uuid-string

        Parameters
        ----------
        uuid
            uuid object, e.g. UUID("2ea5b605-c91f-4e5a-9559-3dff79fdd4a5")

        Returns
        -------
            OSW-ID string, e.g. OSW2ea5b605c91f4e5a95593dff79fdd4a5
        """
        return "OSW" + str(uuid).replace("-", "")

    @staticmethod
    def get_uuid(osw_id) -> uuid:
        """Returns the uuid for a given OSW-ID

        Parameters
        ----------
        osw_id
            OSW-ID string, e.g. OSW2ea5b605c91f4e5a95593dff79fdd4a5

        Returns
        -------
            uuid object, e.g. UUID("2ea5b605-c91f-4e5a-9559-3dff79fdd4a5")
        """
        return UUID(osw_id.replace("OSW", ""))

    class SortEntitiesResult(OswBaseModel):
        by_name: Dict[str, List[OswBaseModel]]
        by_type: Dict[str, List[OswBaseModel]]

    @staticmethod
    def sort_list_of_entities_by_class(
        entities: List[OswBaseModel],
        exclude_typeless: bool = True,
        raise_error: bool = False,
    ) -> SortEntitiesResult:
        """Sorts a list of entities by class name and type.

        Parameters
        ----------
        entities:
            List of entities to be sorted
        exclude_typeless:
            Exclude entities, which are instances of a class that does not
            define a field 'type'
        raise_error:
            Raise an error if an entity can not be processed because it is an
            instance of class that does not define a field 'type'
        """
        by_name = {}
        by_type = {}
        for entity in entities:
            # Get class name
            name = entity.__class__.__name__
            # See if the class has a type field
            if "type" not in entity.__class__.__fields__:
                if raise_error:
                    raise AttributeError(
                        f"Instance '{entity}' of class '{name}' can not be processed "
                        f"as the class does not define a field 'type'."
                    )
                if exclude_typeless:
                    warn(
                        f"Skipping instance '{entity}' of class '{name}' as the class "
                        f"does not define a field 'type'."
                    )
                    # Excludes the respective entity from the list which will be
                    #  processed further:
                    continue
                model_type = None
            else:
                # Get class type if available
                model_type = entity.__class__.__fields__["type"].get_default()[0]
            # Add entity to by_name
            if name not in by_name:
                by_name[name] = []
            by_name[name].append(entity)
            # Add entity to by_type
            if model_type not in by_type:
                by_type[model_type] = []
            by_type[model_type].append(entity)

        return OSW.SortEntitiesResult(by_name=by_name, by_type=by_type)

    class SchemaRegistration(BaseModel):
        """
        dataclass param of register_schema()

        Attributes
        ----------
        model_cls:
            the model class
        schema_name:
            the name of the schema
        schema_bases:
            list of base schemas (referenced by allOf)
        """

        class Config:
            arbitrary_types_allowed = True  # allow any class as type

        model_cls: Type[OswBaseModel]
        schema_uuid: str  # Optional[str] = model_cls.__uuid__
        schema_name: str  # Optional[str] = model_cls.__name__
        schema_bases: List[str] = ["Category:Item"]

    def register_schema(self, schema_registration: SchemaRegistration):
        """Registers a new or updated schema in OSW by creating the corresponding
        category page.

        Parameters
        ----------
        schema_registration
            see SchemaRegistration
        """
        entity = schema_registration.model_cls

        jsondata = {}
        jsondata["uuid"] = schema_registration.schema_uuid
        jsondata["label"] = {"text": schema_registration.schema_name, "lang": "en"}
        jsondata["subclass_of"] = schema_registration.schema_bases

        if issubclass(entity, BaseModel):
            entity_title = "Category:" + OSW.get_osw_id(schema_registration.schema_uuid)

            page = WtPage(wtSite=self.site, title=entity_title)
            if page.exists:
                page = self.site.get_page(
                    WtSite.GetPageParam(titles=[entity_title])
                ).pages[0]

            page.set_slot_content("jsondata", jsondata)

            schema = json.loads(
                entity.schema_json(indent=4).replace("$ref", "dollarref")
            )

            jsonpath_expr = parse("$..allOf")
            # Replace local definitions (#/definitions/...) with embedded definitions
            #  to prevent resolve errors in json-editor
            for match in jsonpath_expr.find(schema):
                result_array = []
                for subschema in match.value:
                    # pprint(subschema)
                    value = subschema["dollarref"]
                    if value.startswith("#"):
                        definition_jsonpath_expr = parse(
                            value.replace("#", "$").replace("/", ".")
                        )
                        for def_match in definition_jsonpath_expr.find(schema):
                            # pprint(def_match.value)
                            result_array.append(def_match.value)
                    else:
                        result_array.append(subschema)
                match.full_path.update_or_create(schema, result_array)
            if "definitions" in schema:
                del schema["definitions"]

            if "allOf" not in schema:
                schema["allOf"] = []
            for base in schema_registration.schema_bases:
                schema["allOf"].append(
                    {"$ref": f"/wiki/{base}?action=raw&slot=jsonschema"}
                )

            page.set_slot_content("jsonschema", schema)
        else:
            print("Error: Unsupported entity type")
            return

        page.edit()
        print("Entity stored at " + page.get_url())

    class SchemaUnregistration(BaseModel):
        """
        dataclass param of register_schema()

        Attributes
        ----------
        model_cls:
            the model class
        schema_name:
            the name of the schema
        schema_bases:
            list of base schemas (referenced by allOf)
        """

        class Config:
            arbitrary_types_allowed = True  # allow any class as type

        model_cls: Optional[Type[OswBaseModel]]
        model_uuid: Optional[str]
        comment: Optional[str]

    def unregister_schema(self, schema_unregistration: SchemaUnregistration):
        """deletes the corresponding category page

        Parameters
        ----------
        schema_unregistration
            see SchemaUnregistration
        """
        uuid = ""
        if schema_unregistration.model_uuid:
            uuid = schema_unregistration.model_uuid
        elif (
            not uuid
            and schema_unregistration.model_cls
            and issubclass(schema_unregistration.model_cls, BaseModel)
        ):
            uuid = schema_unregistration.model_cls.__uuid__
        else:
            print("Error: Neither model nor model id provided")

        entity_title = "Category:" + OSW.get_osw_id(uuid)
        page = self.site.get_page(WtSite.GetPageParam(titles=[entity_title])).pages[0]
        page.delete(schema_unregistration.comment)

    class FetchSchemaMode(Enum):
        """Modes of the FetchSchemaParam class

        Attributes
        ----------
        append:
            append to the current model
        replace:
            replace the current model
        """

        append = "append"  # append to the current model
        replace = "replace"  # replace the current model

    class FetchSchemaParam(BaseModel):
        """Param for fetch_schema()

        Attributes
        ----------
        schema_title:
            one or multiple titles (wiki page name) of schemas (default: Category:Item)
        mode:
            append or replace (default) current schema, see FetchSchemaMode
        """

        schema_title: Optional[Union[List[str], str]] = "Category:Item"
        mode: Optional[str] = (
            "replace"
            # type 'FetchSchemaMode' requires: 'from __future__ import annotations'
        )

    def fetch_schema(self, fetchSchemaParam: FetchSchemaParam = None) -> None:
        """Loads the given schemas from the OSW instance and autogenerates python
        datasclasses within osw.model.entity from it

        Parameters
        ----------
        fetchSchemaParam
            See FetchSchemaParam, by default None
        """
        if not isinstance(fetchSchemaParam.schema_title, list):
            fetchSchemaParam.schema_title = [fetchSchemaParam.schema_title]
        first = True
        for schema_title in fetchSchemaParam.schema_title:
            mode = fetchSchemaParam.mode
            if not first:  # 'replace' makes only sense for the first schema
                mode = "append"
            self._fetch_schema(
                OSW._FetchSchemaParam(schema_title=schema_title, mode=mode)
            )
            first = False

    class _FetchSchemaParam(BaseModel):
        """Internal param for _fetch_schema()

        Attributes
        ----------
        schema_title:
            the title (wiki page name) of the schema (default: Category:Item)
        root:
            marks the root iteration for a recursive fetch (internal param,
            default: True)
        mode:
            append or replace (default) current schema, see FetchSchemaMode
        """

        schema_title: Optional[str] = "Category:Item"
        root: Optional[bool] = True
        mode: Optional[str] = (
            "replace"
            # type 'FetchSchemaMode' requires: 'from __future__ import annotations'
        )

    def _fetch_schema(self, fetchSchemaParam: _FetchSchemaParam = None) -> None:
        """Loads the given schema from the OSW instance and autogenerates python
        datasclasses within osw.model.entity from it

        Parameters
        ----------
        fetchSchemaParam
            See FetchSchemaParam, by default None
        """
        site_cache_state = self.site.get_cache_enabled()
        self.site.enable_cache()
        if fetchSchemaParam is None:
            fetchSchemaParam = OSW._FetchSchemaParam()
        schema_title = fetchSchemaParam.schema_title
        root = fetchSchemaParam.root
        schema_name = schema_title.split(":")[-1]
        page = self.site.get_page(WtSite.GetPageParam(titles=[schema_title])).pages[0]
        if not page.exists:
            print(f"Error: Page {schema_title} does not exist")
            return
        # not only in the JsonSchema namespace the schema is located in the main sot
        # in all other namespaces, the json_schema slot is used
        if schema_title.startswith("JsonSchema:"):
            schema_str = ""
            if page.get_slot_content("main"):
                schema_str = json.dumps(page.get_slot_content("main"))
        else:
            schema_str = ""
            if page.get_slot_content("jsonschema"):
                schema_str = json.dumps(page.get_slot_content("jsonschema"))
        if (schema_str is None) or (schema_str == ""):
            print(f"Error: Schema {schema_title} does not exist")
            schema_str = "{}"  # empty schema to make reference work
        schema = json.loads(
            schema_str.replace("$ref", "dollarref").replace(
                # '$' is a special char for root object in jsonpath
                '"allOf": [',
                '"allOf": [{},',
            )
            # fix https://github.com/koxudaxi/datamodel-code-generator/issues/1910
        )
        print(f"Fetch {schema_title}")

        jsonpath_expr = parse("$..dollarref")
        for match in jsonpath_expr.find(schema):
            # value = "https://" + self.mw_site.host + match.value
            if match.value.startswith("#"):
                continue  # skip self references
            ref_schema_title = match.value.replace("/wiki/", "").split("?")[0]
            ref_schema_name = ref_schema_title.split(":")[-1] + ".json"
            value = ""
            for _i in range(0, schema_name.count("/")):
                value += "../"  # created relative path to top-level schema dir
            value += ref_schema_name  # create a reference to a local file
            # keep document-relative jsonpointer if present
            if "#/" in match.value:
                value += "#/" + match.value.split("#/")[-1]
            match.full_path.update_or_create(schema, value)
            # print(f"replace {match.value} with {value}")
            if (
                ref_schema_title != schema_title
            ):  # prevent recursion in case of self references
                self._fetch_schema(
                    OSW._FetchSchemaParam(schema_title=ref_schema_title, root=False)
                )  # resolve references recursive

        model_dir_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "model"
        )  # src/model
        schema_path = os.path.join(model_dir_path, schema_name + ".json")
        os.makedirs(os.path.dirname(schema_path), exist_ok=True)
        with open(schema_path, "w", encoding="utf-8") as f:
            schema_str = json.dumps(schema, ensure_ascii=False, indent=4).replace(
                "dollarref", "$ref"
            )
            # print(schema_str)
            f.write(schema_str)

        # result_model_path = schema_path.replace(".json", ".py")
        result_model_path = os.path.join(model_dir_path, "entity.py")
        temp_model_path = os.path.join(model_dir_path, "temp.py")
        if root:
            exec_name = "datamodel-codegen"
            # default: assume datamodel-codegen is in PATH
            exec_path = exec_name
            if platform.system() == "Windows":
                exec_name += ".exe"
                exec_path = os.path.join(
                    os.path.dirname(os.path.abspath(sys.executable)), exec_name
                )
                if not os.path.isfile(exec_path):
                    exec_path = os.path.join(
                        os.path.dirname(os.path.abspath(sys.executable)),
                        "Scripts",
                        exec_name,
                    )
                if not os.path.isfile(exec_path):
                    print("Error: datamodel-codegen not found")
                    return
            os.system(
                f"{exec_path}  \
                --input {schema_path} \
                --input-file-type jsonschema \
                --output {temp_model_path} \
                --base-class osw.model.static.OswBaseModel \
                --use-default \
                --use-unique-items-as-set \
                --enum-field-as-literal all \
                --use-title-as-name \
                --use-schema-description \
                --use-field-description \
                --encoding utf-8 \
                --use-double-quotes \
                --collapse-root-models \
                --reuse-model \
            "
            )
            # see https://koxudaxi.github.io/datamodel-code-generator/
            # --base-class OswBaseModel: use a custom base class
            # --custom-template-dir src/model/template_data/
            # --extra-template-data src/model/template_data/extra.json
            # --use-default: Use default value even if a field is required
            # --use-unique-items-as-set: define field type as `set` when the field
            #  attribute has`uniqueItems`
            # --enum-field-as-literal all: prevent 'value is not a valid enumeration
            #  member' errors after schema reloading
            # --use-schema-description: Use schema description to populate class
            #  docstring
            # --use-field-description: Use schema description to populate field
            #  docstring
            # --use-title-as-name: use titles as class names of models, e.g. for the
            #  footer templates
            # --collapse-root-models: Models generated with a root-type field will be
            #  merged
            # into the models using that root-type model, e.g. for Entity.statements
            # --reuse-model: Re-use models on the field when a module has the model
            #  with the same content

            content = ""
            with open(temp_model_path, "r", encoding="utf-8") as f:
                content = f.read()
            os.remove(temp_model_path)

            content = re.sub(
                r"(UUID = Field\(...)",
                r"UUID = Field(default_factory=uuid4",
                content,
            )  # enable default value for uuid

            # we are now using pydantic.v1
            # pydantic imports lead to uninitialized fields (FieldInfo still present)
            content = re.sub(
                r"(from pydantic import)", "from pydantic.v1 import", content
            )

            # remove field param unique_items
            # --use-unique-items-as-set still keeps unique_items=True as Field param
            # which was removed, see https://github.com/pydantic/pydantic-core/issues/296
            # --output-model-type pydantic_v2.BaseModel fixes that but generated models
            # are not v1 compatible mainly by using update_model()
            content = re.sub(r"(,?\s*unique_items=True\s*)", "", content)

            if fetchSchemaParam.mode == "replace":
                header = (
                    "from uuid import uuid4\n"
                    "from typing import Type, TypeVar\n"
                    "from osw.model.static import OswBaseModel, Ontology\n"
                    # "from osw.model.static import *\n"
                    "\n"
                )

                content = re.sub(
                    pattern=r"(class\s*\S*\s*\(\s*OswBaseModel\s*\)\s*:.*\n)",
                    repl=header + r"\n\n\n\1",
                    string=content,
                    count=1,
                )  # add header before first class declaration

                with open(result_model_path, "w", encoding="utf-8") as f:
                    f.write(content)

            if fetchSchemaParam.mode == "append":
                org_content = ""
                with open(result_model_path, "r", encoding="utf-8") as f:
                    org_content = f.read()

                pattern = re.compile(
                    r"class\s*([\S]*)\s*\(\s*\S*\s*\)\s*:.*\n"
                )  # match class definition [\s\S]*(?:[^\S\n]*\n){2,}
                for cls in re.findall(pattern, org_content):
                    print(cls)
                    content = re.sub(
                        r"(class\s*"
                        + cls
                        + r"\s*\(\s*\S*\s*\)\s*:.*\n[\s\S]*?(?:[^\S\n]*\n){3,})",
                        "",
                        content,
                        count=1,
                    )  # replace duplicated classes

                content = re.sub(
                    pattern=r"(from __future__ import annotations)",
                    repl="",
                    string=content,
                    count=1,
                )  # remove import statement
                # print(content)
                with open(result_model_path, "a", encoding="utf-8") as f:
                    f.write(content)

            importlib.reload(model)  # reload the updated module
            if not site_cache_state:
                self.site.disable_cache()  # restore original state

    class LoadEntityParam(BaseModel):
        """Param for load_entity()"""

        titles: Union[str, List[str]]
        """The pages titles to load - one or multiple titles (wiki page name) of
        entities"""
        autofetch_schema: Optional[bool] = True
        """If true, load the corresponding schemas /
        categories ad-hoc if not already present"""
        remove_empty: Optional[bool] = True
        """If true, remove key with an empty string, list, dict or set as value
        from the jsondata."""
        disable_cache: bool = False
        """If true, disable the cache for the loading process"""

        def __init__(self, **data):
            super().__init__(**data)
            if not isinstance(self.titles, list):
                self.titles = [self.titles]

    class LoadEntityResult(BaseModel):
        """Result of load_entity()"""

        entities: Union[model.Entity, List[model.Entity]]
        """The dataclass instance(s)"""

    def load_entity(
        self, entity_title: Union[str, List[str], LoadEntityParam]
    ) -> Union[model.Entity, List[model.Entity], LoadEntityResult]:
        """Loads the entity with the given wiki page name from the OSW instance.
        Creates an instance of the class specified by the "type" attribute, default
        model.Entity. An instance of model.Entity can be cast to any subclass with
        .cast(model.<class>) .

        Parameters
        ----------
        entity_title
            the wiki page name

        Returns
        -------
            the dataclass instance if only a single title is given
            a list of dataclass instances if a list of titles is given
            a LoadEntityResult instance if a LoadEntityParam is given
        """
        if isinstance(entity_title, str):
            param = OSW.LoadEntityParam(titles=[entity_title])
        elif isinstance(entity_title, list):
            param = OSW.LoadEntityParam(titles=entity_title)
        else:
            param = entity_title

        # store original cache state
        cache_state = self.site.get_cache_enabled()
        if param.disable_cache:
            self.site.disable_cache()
        if not cache_state and param.disable_cache:
            # enable cache to speed up loading
            self.site.enable_cache()

        entities = []
        pages = self.site.get_page(WtSite.GetPageParam(titles=param.titles)).pages
        for page in pages:
            entity = None
            schemas = []
            schemas_fetched = True
            jsondata = page.get_slot_content("jsondata")
            if param.remove_empty:
                remove_empty(jsondata)
            if jsondata:
                for category in jsondata["type"]:
                    schema = (
                        self.site.get_page(WtSite.GetPageParam(titles=[category]))
                        .pages[0]
                        .get_slot_content("jsonschema")
                    )
                    schemas.append(schema)
                    # generate model if not already exists
                    cls_name: str = schema["title"]
                    if not hasattr(model, cls_name):
                        if param.autofetch_schema:
                            self.fetch_schema(
                                OSW.FetchSchemaParam(
                                    schema_title=category, mode="append"
                                )
                            )
                    if not hasattr(model, cls_name):
                        schemas_fetched = False
                        print(
                            f"Error: Model {cls_name} not found. Schema {category} "
                            f"needs to be fetched first."
                        )
            if not schemas_fetched:
                continue

            if len(schemas) == 0:
                print("Error: no schema defined")

            elif len(schemas) == 1:
                cls: Type[model.Entity] = getattr(model, schemas[0]["title"])
                entity: model.Entity = cls(**jsondata)

            else:
                bases = []
                for schema in schemas:
                    bases.append(getattr(model, schema["title"]))
                cls = create_model("Test", __base__=tuple(bases))
                entity: model.Entity = cls(**jsondata)

            if entity is not None:
                # make sure we do not override existing metadata
                if not hasattr(entity, "meta") or entity.meta is None:
                    entity.meta = model.Meta()
                if (
                    not hasattr(entity.meta, "wiki_page")
                    or entity.meta.wiki_page is None
                ):
                    entity.meta.wiki_page = model.WikiPage()
                entity.meta.wiki_page.namespace = namespace_from_full_title(page.title)
                entity.meta.wiki_page.title = title_from_full_title(page.title)

            entities.append(entity)
        # restore original cache state
        if cache_state:
            self.site.enable_cache()
        else:
            self.site.disable_cache()

        if isinstance(entity_title, str):  # single title
            if len(entities) >= 1:
                return entities[0]
            else:
                return None
        if isinstance(entity_title, list):  # list of titles
            return entities
        if isinstance(entity_title, OSW.LoadEntityParam):  # LoadEntityParam
            return OSW.LoadEntityResult(entities=entities)

    class OverwriteClassParam(OswBaseModel):
        model: Type[OswBaseModel]  # ModelMetaclass
        """The model class for which this is the overwrite params object."""
        overwrite: Optional[OVERWRITE_CLASS_OPTIONS] = False
        """Defines the overall overwriting behavior. Used for any property if the
        property specific setting is not set."""
        per_property: Optional[Dict[str, OverwriteOptions]] = None
        """A key (property name) - value (overwrite setting) pair."""
        _per_property: Dict[str, OVERWRITE_CLASS_OPTIONS] = PrivateAttr()
        """Private property, for internal use only. Use 'per_property' instead"""

        @validator("per_property")
        def validate_per_property(cls, per_property, values):
            model_ = values.get("model")
            field_names = list(model_.__fields__.keys())
            keys = per_property.keys()
            if not all(key in field_names for key in keys):
                missing_keys = [key for key in keys if key not in field_names]
                raise ValueError(
                    f"Property not found in model: {', '.join(missing_keys)}"
                )

            return per_property

        def __setattr__(self, key, value):
            """Called when setting an attribute"""
            super().__setattr__(key, value)
            if key == "per_property":
                # compare value and self.per_property
                if value != self.per_property and value is not None:
                    self._per_property = {
                        field_name: value.get(field_name, self.overwrite)
                        for field_name in self.model.__fields__.keys()
                    }
            elif key == "overwrite":
                if self.per_property is not None:
                    self._per_property = {
                        field_name: self.per_property.get(field_name, self.overwrite)
                        for field_name in self.model.__fields__.keys()
                    }
            elif key == "model":
                if self.per_property is not None:
                    self._per_property = {
                        field_name: self.per_property.get(field_name, self.overwrite)
                        for field_name in self.model.__fields__.keys()
                    }

        def __init__(self, **data):
            """Called after validation. Sets the fallback for every property that
            has not been specified in per_property."""
            super().__init__(**data)
            per_property_ = {}
            if self.per_property is not None:
                per_property_ = self.per_property
            self._per_property = {
                field_name: per_property_.get(field_name, self.overwrite)
                for field_name in self.model.__fields__.keys()
            }
            # todo: from class definition get properties with hidden /
            #  read_only option  #  those can be safely overwritten - set the to True

        def get_overwrite_setting(self, property_name: str) -> OverwriteOptions:
            """Returns the fallback overwrite option for the given field name"""
            return self._per_property.get(property_name, self.overwrite)

    class _ApplyOverwriteParam(OswBaseModel):
        page: WtPage
        entity: OswBaseModel  # actually model.Entity but this causes the "type" error
        policy: Union[OSW.OverwriteClassParam, OVERWRITE_CLASS_OPTIONS]
        namespace: Optional[str]
        meta_category_title: Optional[str]
        meta_category_template_str: Optional[str]
        remove_empty: Optional[bool] = True
        inplace: Optional[bool] = False
        debug: Optional[bool] = False

        class Config:
            arbitrary_types_allowed = True

        @validator("entity")
        def validate_entity(cls, entity, values):
            """Make sure that the passed entity has the same uuid as the page"""
            page: WtPage = values.get("page")
            if not page.exists:  # Guard clause
                return entity
            jsondata = page.get_slot_content("jsondata")
            if jsondata is None:  # Guard clause
                title = title_from_full_title(page.title)
                try:
                    uuid_from_title = get_uuid(title)
                except ValueError:
                    print(
                        f"Error: UUID could not be determined from title: '{title}', "
                        f"nor fromjsondata: {jsondata}"
                    )
                    return entity
                if str(uuid_from_title) != str(entity.uuid):
                    raise ValueError(
                        f"UUID mismatch: Page UUID: {uuid_from_title}, "
                        f"Entity UUID: {entity.uuid}"
                    )
                return entity
            page_uuid = str(jsondata.get("uuid"))
            entity_uuid = str(getattr(entity, "uuid", None))
            if page_uuid != entity_uuid or page_uuid == str(None):
                # Comparing string type UUIDs
                raise ValueError(
                    f"UUID mismatch: Page UUID: {page_uuid}, Entity UUID: {entity_uuid}"
                )
            return entity

        def __init__(self, **data):
            super().__init__(**data)
            if self.namespace is None:
                self.namespace = get_namespace(self.entity)
            if self.namespace is None:
                raise ValueError("Namespace could not be determined.")
            if not isinstance(self.policy, OSW.OverwriteClassParam):
                self.policy = OSW.OverwriteClassParam(
                    model=self.entity.__class__,
                    overwrite=self.policy,
                )

    @staticmethod
    def _apply_overwrite_policy(param: OSW._ApplyOverwriteParam) -> WtPage:
        if param.inplace:
            page = param.page
        else:
            page = deepcopy(param.page)
        entity_title = f"{param.namespace}:{get_title(param.entity)}"

        def set_content(content_to_set: dict) -> None:
            if param.debug:
                print(f"content_to_set: {str(content_to_set)}")
            for slot_ in content_to_set.keys():
                page.set_slot_content(slot_, content_to_set[slot_])

        # Create a variable to hold the new content
        new_content = {
            # required for json parsing and header rendering
            "header": "{{#invoke:Entity|header}}",
            # required for footer rendering
            "footer": "{{#invoke:Entity|footer}}",
        }
        # Take the shortcut if
        # 1. page does not exist AND any setting of overwrite
        # 2. overwrite is "replace remote"
        if (
            not page.exists
            or param.policy.overwrite == AddOverwriteClassOptions.replace_remote
        ):
            # Use pydantic serialization, skip none values:
            new_content["jsondata"] = json.loads(param.entity.json(exclude_none=True))
            if param.remove_empty:
                remove_empty(new_content["jsondata"])
            set_content(new_content)
            page.changed = True
            return page  # Guard clause --> exit function
        # 3. pages does exist AND overwrite is "keep existing"
        if (
            page.exists
            and param.policy.overwrite == AddOverwriteClassOptions.keep_existing
        ):
            print(
                f"Entity '{entity_title}' already exists and won't be stored "
                f"with overwrite set to 'keep existing'!"
            )
            return page  # Guard clause --> exit function
        # Apply the overwrite logic in any other case
        # 4. If per_property was None  -> overwrite will be used as a fallback
        # 4.1 If overwrite is True ---> overwrite existing properties
        # 4.2 If overwrite is False --> don't overwrite existing properties
        # 4.3 If overwrite is "only empty" --> overwrite existing properties if
        #     they are empty
        # * Download page
        # * Merge slots selectively based on per_property

        # Create variables to hold local and remote content prior to merging
        local_content = {}
        remote_content = {}
        # Get the remote content
        for slot in ["jsondata", "header", "footer"]:  # SLOTS:
            remote_content[slot] = page.get_slot_content(slot)
            # Todo: remote content does not contain properties that are not set
        if param.remove_empty:
            remove_empty(remote_content["jsondata"])
        if remote_content["header"]:  # not None or {} or ""
            new_content["header"] = remote_content["header"]
        if remote_content["footer"]:
            new_content["footer"] = remote_content["footer"]
        if param.debug:
            print(f"'remote_content': {str(remote_content)}")
        # Get the local content
        # Properties that are not set in the local content will be set to None
        # We want those not to be listed as keys
        local_content["jsondata"] = json.loads(param.entity.json(exclude_none=True))
        if param.remove_empty:
            remove_empty(local_content["jsondata"])
        if param.debug:
            print(f"'local_content': {str(remote_content)}")
        # Apply the overwrite logic
        # a) If there is a key in the remote content that is not in the local
        #    content, we have to keep it
        new_content["jsondata"] = remote_content["jsondata"]
        # new_content["jsondata"] = {
        #     key: value
        #     for (key, value) in remote_content["jsondata"].items()
        #     if key not in local_content["jsondata"].keys()
        # }
        if param.debug:
            print(f"'New content' after 'remote' update: {str(new_content)}")
        # b) If there is a key in the local content that is not in the remote
        #    content, we have to keep it
        new_content["jsondata"].update(
            {
                key: value
                for (key, value) in local_content["jsondata"].items()
                if key not in remote_content["jsondata"].keys()
            }
        )
        if param.debug:
            print(f"'New content' after 'local' update: {str(new_content)}")
        # c) If there is a key in both contents, we have to apply the overwrite
        #    logic
        # todo: include logic for hidden and read_only properties!
        new_content["jsondata"].update(
            {
                key: value
                for (key, value) in local_content["jsondata"].items()
                if param.policy.get_overwrite_setting(key) == OverwriteOptions.true
            }
        )
        if param.debug:
            print(f"'New content' after 'True' update: {str(new_content)}")
        new_content["jsondata"].update(
            {
                key: value
                for (key, value) in remote_content["jsondata"].items()
                if param.policy.get_overwrite_setting(key) == OverwriteOptions.false
            }
        )
        if param.debug:
            print(f"'New content' after 'False' update: {str(new_content)}")
        new_content["jsondata"].update(
            {
                key: value
                for (key, value) in local_content["jsondata"].items()
                if (
                    param.policy.get_overwrite_setting(key)
                    == OverwriteOptions.only_empty
                    and is_empty(remote_content["jsondata"].get(key))
                )
            }
        )
        if param.debug:
            print(f"'New content' after 'only empty' update: {str(new_content)}")
            print(f"'New content' to be stored: {str(new_content)}")
        set_content(new_content)
        return page  # Guard clause --> exit function

    class StoreEntityParam(OswBaseModel):
        entities: Union[OswBaseModel, List[OswBaseModel]]  # actually model.Entity
        """The entities to store. Can be a single entity or a list of entities."""
        namespace: Optional[str]
        """The namespace of the entities. If not set, the namespace is derived from the
        entity."""
        parallel: Optional[bool] = None
        """If set to True, the entities are stored in parallel."""
        overwrite: Optional[OVERWRITE_CLASS_OPTIONS] = "keep existing"
        """If no class specific overwrite setting is set, this setting is used."""
        overwrite_per_class: Optional[List[OSW.OverwriteClassParam]] = None
        """A list of OverwriteClassParam objects. If a class specific overwrite setting
        is set, this setting is used.
        """
        remove_empty: Optional[bool] = True
        """If true, remove key with an empty string value from the jsondata."""
        change_id: Optional[str] = None
        """ID to document the change. Entities within the same store_entity() call will
        share the same change_id. This parameter can also be used to link multiple
        store_entity() calls."""
        meta_category_title: Optional[str] = "Category:Category"
        debug: Optional[bool] = False
        _overwrite_per_class: Dict[str, Dict[str, OSW.OverwriteClassParam]] = (
            PrivateAttr()
        )
        """Private attribute, for internal use only. Use 'overwrite_per_class'
        instead."""

        def __init__(self, **data):
            super().__init__(**data)
            if not isinstance(self.entities, list):
                self.entities = [self.entities]
            if self.change_id is None:
                self.change_id = str(uuid4())
            for entity in self.entities:
                if getattr(entity, "meta", None) is None:
                    entity.meta = model.Meta()
                if entity.meta.change_id is None:
                    entity.meta.change_id = []
                if self.change_id not in entity.meta.change_id:
                    entity.meta.change_id.append(self.change_id)
            if len(self.entities) > 5 and self.parallel is None:
                self.parallel = True
            if self.parallel is None:
                self.parallel = (
                    True  # Set to True after implementation of asynchronous upload
                )
            if self.overwrite is None:
                self.overwrite = self.__fields__["overwrite"].get_default()
            self._overwrite_per_class = {"by name": {}, "by type": {}}
            if self.overwrite_per_class is not None:
                for param in self.overwrite_per_class:
                    model_name = param.model.__name__
                    model_type = param.model.__fields__["type"].get_default()[0]
                    if (
                        model_name in self._overwrite_per_class["by name"].keys()
                        or model_type in self._overwrite_per_class["by type"].keys()
                    ):
                        raise ValueError(
                            f"More than one OverwriteClassParam for the class "
                            f"'{model_type}' ({model_name}) has been passed in the "
                            f"list to 'overwrite_per_class'!"
                        )
                    self._overwrite_per_class["by name"][model_name] = param
                    self._overwrite_per_class["by type"][model_type] = param

    class StoreEntityResult(OswBaseModel):
        """Result of store_entity()"""

        change_id: str
        """The ID of the change"""

    def store_entity(
        self, param: Union[StoreEntityParam, OswBaseModel, List[OswBaseModel]]
    ) -> StoreEntityResult:
        """stores the given dataclass instance as OSW page by calling BaseModel.json()

        Parameters
        ----------
        param:
            StoreEntityParam, the dataclass instance or a list of instances
        """
        if isinstance(param, model.Entity):
            param = OSW.StoreEntityParam(entities=[param])
        if isinstance(param, list):
            param = OSW.StoreEntityParam(entities=param)
        if not isinstance(param.entities, list):
            param.entities = [param.entities]

        param: OSW.StoreEntityParam = param

        max_index = len(param.entities)

        meta_category = self.site.get_page(
            WtSite.GetPageParam(titles=[param.meta_category_title])
        ).pages[0]
        # ToDo: we have to do this iteratively to support meta categories inheritance
        meta_category_template_str = meta_category.get_slot_content("schema_template")
        meta_category_template = None
        if param.namespace == "Category":
            if param.meta_category_title:
                meta_category = self.site.get_page(
                    WtSite.GetPageParam(titles=[param.meta_category_title])
                ).pages[0]
                meta_category_template_str = meta_category.get_slot_content(
                    "schema_template"
                )
            if meta_category_template_str:
                meta_category_template = compile_handlebars_template(
                    meta_category_template_str
                )

        def store_entity_(
            entity_: model.Entity,
            namespace_: str = None,
            index: int = None,
            overwrite_class_param: OSW.OverwriteClassParam = None,
        ) -> None:
            title_ = get_title(entity_)
            if namespace_ is None:
                namespace_ = get_namespace(entity_)
            if namespace_ is None or title_ is None:
                print("Error: Unsupported entity type")
                return
            if overwrite_class_param is None:
                raise TypeError("'overwrite_class_param' must not be None!")
            entity_title = namespace_ + ":" + title_
            page = self._apply_overwrite_policy(
                OSW._ApplyOverwriteParam(
                    page=WtPage(wtSite=self.site, title=entity_title),
                    entity=entity_,
                    namespace=namespace_,
                    policy=overwrite_class_param,
                    meta_category_template_str=meta_category_template_str,
                    remove_empty=param.remove_empty,
                    debug=param.debug,
                )
            )
            if meta_category_template:
                try:
                    jsondata = page.get_slot_content("jsondata")
                    if param.remove_empty:
                        remove_empty(jsondata)
                    schema_str = eval_compiled_handlebars_template(
                        meta_category_template,
                        jsondata,
                        {
                            "_page_title": entity_title,  # Legacy
                            "_current_subject_": entity_title,
                        },
                    )
                    schema = json.loads(schema_str)
                    # Put generated schema in definitions section,
                    #  currently only enabled for Characteristics
                    if hasattr(model, "CharacteristicType") and isinstance(
                        entity_, model.CharacteristicType
                    ):
                        new_schema = {
                            "$defs": {"generated": schema},
                            "allOf": [{"$ref": "#/$defs/generated"}],
                            "@context": schema.pop("@context", None),
                            "title": schema.pop("title", ""),
                        }
                        schema["title"] = "Generated" + new_schema["title"]
                        schema = new_schema
                    page.set_slot_content("jsonschema", new_schema)
                except Exception as e:
                    print(f"Schema generation from template failed for {entity_}: {e}")
            page.edit()  # will set page.changed if the content of the page has changed
            if page.changed:
                if index is None:
                    print(f"Entity stored at '{page.get_url()}'.")
                else:
                    print(
                        f"({index + 1}/{max_index}) Entity stored at "
                        f"'{page.get_url()}'."
                    )

        sorted_entities = OSW.sort_list_of_entities_by_class(param.entities)
        print(
            "Entities to be uploaded have been sorted according to their type.\n"
            "If you would like to overwrite existing entities or properties, "
            "pass a StoreEntityParam to store_entity() with "
            "attribute 'overwrite' or 'overwrite_per_class' set to, e.g., "
            "True."
        )

        class UploadObject(BaseModel):
            entity: OswBaseModel
            # Actually model.Entity but this causes the "type" error
            namespace: Optional[str]
            index: int
            overwrite_class_param: OSW.OverwriteClassParam

        upload_object_list: List[UploadObject] = []

        upload_index = 0
        for class_type, entities in sorted_entities.by_type.items():
            # Try to get a class specific overwrite setting
            class_param = param._overwrite_per_class["by type"].get(class_type, None)
            if class_param is None:
                entity_model = entities[0].__class__
                class_param = OSW.OverwriteClassParam(
                    model=entity_model,
                    overwrite=param.overwrite,
                )
                if param.debug:
                    print(
                        f"Now adding entities of class type '{class_type}' "
                        f"({entity_model.__name__}) to upload list. No class specific"
                        f" overwrite setting found. Using fallback option '"
                        f"{param.overwrite}' for all entities of this class."
                    )
            for entity in entities:
                upload_object_list.append(
                    UploadObject(
                        entity=entity,
                        namespace=param.namespace,
                        index=upload_index,
                        overwrite_class_param=class_param,
                    )
                )
                upload_index += 1

        def handle_upload_object_(upload_object: UploadObject) -> None:
            store_entity_(
                upload_object.entity,
                upload_object.namespace,
                upload_object.index,
                upload_object.overwrite_class_param,
            )

        if param.parallel:
            _ = parallelize(
                handle_upload_object_, upload_object_list, flush_at_end=param.debug
            )
        else:
            _ = [
                handle_upload_object_(upload_object)
                for upload_object in upload_object_list
            ]
        return OSW.StoreEntityResult(change_id=param.change_id)

    class DeleteEntityParam(OswBaseModel):
        entities: Union[OswBaseModel, List[OswBaseModel]]
        comment: Optional[str] = None
        parallel: Optional[bool] = None
        debug: Optional[bool] = False

        def __init__(self, **data):
            super().__init__(**data)
            if not isinstance(self.entities, list):
                self.entities = [self.entities]
            if len(self.entities) > 5 and self.parallel is None:
                self.parallel = True
            if self.parallel is None:
                self.parallel = False

    def delete_entity(
        self,
        entity: Union[OswBaseModel, List[OswBaseModel], DeleteEntityParam],
        comment: str = None,
    ):
        """Deletes the given entity/entities from the OSW instance."""
        if not isinstance(entity, OSW.DeleteEntityParam):
            entity = OSW.DeleteEntityParam(entities=entity)
        if comment is not None:
            entity.comment = comment

        def delete_entity_(entity_, comment_: str = None):
            """Deletes the given entity from the OSW instance.

            Parameters
            ----------
            entity_:
                The dataclass instance to delete
            comment_:
                Command for the change log, by default None
            """
            title_ = None
            namespace_ = None
            if hasattr(entity_, "meta"):
                if entity_.meta and entity_.meta.wiki_page:
                    if entity_.meta.wiki_page.title:
                        title_ = entity_.meta.wiki_page.title
                    if entity_.meta.wiki_page.namespace:
                        namespace_ = entity_.meta.wiki_page.namespace
            if namespace_ is None:
                namespace_ = get_namespace(entity_)
            if title_ is None:
                title_ = OSW.get_osw_id(entity_.uuid)
            if namespace_ is None or title_ is None:
                print("Error: Unsupported entity type")
                return
            entity_title = namespace_ + ":" + title_
            page = self.site.get_page(WtSite.GetPageParam(titles=[entity_title])).pages[
                0
            ]

            if page.exists:
                page.delete(comment_)
                print("Entity deleted: " + page.get_url())
            else:
                print(f"Entity '{entity_title}' does not exist!")

        if entity.parallel:
            _ = parallelize(
                delete_entity_,
                entity.entities,
                flush_at_end=entity.debug,
                comment_=entity.comment,
            )
        else:
            _ = [delete_entity_(e, entity.comment) for e in entity.entities]

    class QueryInstancesParam(OswBaseModel):
        categories: Union[
            Union[str, Type[OswBaseModel]], List[Union[str, Type[OswBaseModel]]]
        ]
        parallel: Optional[bool] = None
        debug: Optional[bool] = False
        limit: Optional[int] = 1000
        _category_string_parts: List[Dict[str, str]] = PrivateAttr()
        _titles: List[str] = PrivateAttr()

        @staticmethod
        def get_full_page_name_parts(
            category_: Union[str, Type[OswBaseModel]]
        ) -> Dict[str, str]:
            error_msg = (
                f"Category must be a string like 'Category:<category name>' or a "
                f"dataclass subclass with a 'type' attribute. This error occurred on "
                f"'{str(category_)}'"
            )
            if isinstance(category_, str):
                string_to_split = category_
            elif issubclass(category_, OswBaseModel):
                type_ = category_.__fields__.get("type")
                if getattr(type_, "default", None) is None:
                    raise TypeError(error_msg)
                string_to_split = type_.default[0]
            else:
                raise TypeError(error_msg)
            if "Category:" not in string_to_split:
                raise TypeError(error_msg)
            return {
                "namespace": string_to_split.split(":")[0],
                "title": string_to_split.split(":")[-1],
            }

        def __init__(self, **data):
            super().__init__(**data)
            if not isinstance(self.categories, list):
                self.categories = [self.categories]
            if len(self.categories) > 5 and self.parallel is None:
                self.parallel = True
            if self.parallel is None:
                self.parallel = False
            self._category_string_parts = [
                OSW.QueryInstancesParam.get_full_page_name_parts(cat)
                for cat in self.categories
            ]
            self._titles = [parts["title"] for parts in self._category_string_parts]

    def query_instances(
        self, category: Union[str, OswBaseModel, OSW.QueryInstancesParam]
    ) -> List[str]:
        if not isinstance(category, OSW.QueryInstancesParam):
            category = OSW.QueryInstancesParam(categories=category)
        page_titles = category._titles
        search_param = SearchParam(
            query=[f"[[HasType::Category:{page_title}]]" for page_title in page_titles],
            **category.dict(
                exclude={"categories", "_category_string_parts", "_titles"}
            ),
        )
        full_page_titles = self.site.semantic_search(search_param)
        return full_page_titles

    class JsonLdMode(str, Enum):
        """enum for jsonld processing mode"""

        expand = "expand"
        flatten = "flatten"
        compact = "compact"
        frame = "frame"

    class ExportJsonLdParams(OswBaseModel):
        context_loader_config: Optional[WtSite.JsonLdContextLoaderParams] = None
        """The configuration for the JSON-LD context loader."""
        entities: Union[OswBaseModel, List[OswBaseModel]]
        """The entities to convert to JSON-LD. Can be a single entity or a list of
        entities."""
        id_keys: Optional[List[str]] = ["osw_id"]
        """The keys to use as @id in the JSON-LD output. If not found in the entity at root
        level, the full page title is used."""
        resolve_context: Optional[bool] = True
        """If True, remote context URLs are resolved."""
        mode: Optional[OSW.JsonLdMode] = "expand"
        """The JSON-LD processing mode to apply if resolve_context is True."""
        context: Optional[Union[str, list, Dict[str, Any]]] = None
        """The JSON-LD context to apply. Replaces any existing context."""
        additional_context: Optional[Union[str, list, Dict[str, Any]]] = None
        """The JSON-LD context to apply on top of the existing context."""
        frame: Optional[Dict[str, Any]] = None
        """The JSON-LD frame to use for framed mode. If not set, the existing context is used"""
        build_rdf_graph: Optional[bool] = False
        """If True, the output is a graph."""
        debug: Optional[bool] = False

        def __init__(self, **data):
            super().__init__(**data)
            if not isinstance(self.entities, list):
                self.entities = [self.entities]

    class ExportJsonLdResult(OswBaseModel):
        documents: List[Union[Dict[str, Any]]]
        """A single JSON-LD document per entity"""
        graph_document: Dict[str, Any] = None
        """A single JSON-LD document with a @graph element containing all entities"""
        graph: rdflib.Graph = None
        """RDF graph containing all entities. Build only if build_rdf_graph is True"""

        class Config:
            arbitrary_types_allowed = True

    def export_jsonld(self, params: ExportJsonLdParams) -> ExportJsonLdResult:
        """Exports the given entity/entities as JSON-LD."""

        if params.resolve_context:
            jsonld.set_document_loader(
                self.site.get_jsonld_context_loader(params.context_loader_config)
            )

        documents = []
        graph_document = {"@graph": []}
        graph = None
        if params.build_rdf_graph:
            graph = rdflib.Graph()
            prefixes = self.site.get_prefix_dict()
            for prefix in prefixes:
                graph.bind(prefix, prefixes[prefix])

        for e in params.entities:
            data = json.loads(e.json(exclude_none=True, indent=4, ensure_ascii=False))

            data["@context"] = []
            if params.id_keys is not None:
                # append "@id" mappings to the context in an additional object
                id_mapping = {}
                for k in params.id_keys:
                    id_mapping[k] = "@id"
                data["@context"].append(id_mapping)
            if params.context is None:
                for t in e.type:
                    data["@context"].append("/wiki/" + t)
                if params.context is not None:
                    data["@context"].append(params.context)
            else:
                data["@context"].append(self.site.get_jsonld_context_prefixes())
                if isinstance(params.context, list):
                    data["@context"].extend(params.context)
                else:
                    data["@context"].append(params.context)
            if params.additional_context is not None:
                if data["@context"] is None:
                    data["@context"] = []
                elif not isinstance(data["@context"], list):
                    data["@context"] = [data["@context"]]
                data["@context"].append(params.additional_context)

            # if none of the id_keys is found, use the full title
            if not any(k in data for k in params.id_keys):
                data["@id"] = get_full_title(e)

            if params.resolve_context:
                graph_document["@graph"].append(jsonld.expand(data))
                if params.mode == "expand":
                    data = jsonld.expand(data)
                    if isinstance(data, list) and len(data) > 0:
                        data = data[0]
                elif params.mode == "flatten":
                    data = jsonld.flatten(data)
                elif params.mode == "compact":
                    # data = jsonld.expand(data)
                    # if isinstance(data, list): data = data[0]
                    data = jsonld.compact(
                        data,
                        data["@context"] if params.context is None else params.context,
                    )
                elif params.mode == "frame":
                    data = jsonld.frame(
                        data,
                        (
                            {"@context": data["@context"]}
                            if params.frame is None
                            else params.frame
                        ),
                    )

                if params.build_rdf_graph:
                    graph.parse(data=json.dumps(data), format="json-ld")

            documents.append(data)

        result = OSW.ExportJsonLdResult(
            documents=documents, graph_document=graph_document, graph=graph
        )
        return result


OSW._ApplyOverwriteParam.update_forward_refs()
OSW.StoreEntityParam.update_forward_refs()
OSW.ExportJsonLdParams.update_forward_refs()
